"""Entrypoint to generate reference docs for the Anyscale CLI/SDK.

Usage: python -m anyscale._private.docgen --help
"""

import os

import click

import anyscale
from anyscale import scripts
from anyscale._private.docgen.generator import MarkdownGenerator, Module
from anyscale.aggregated_instance_usage.models import DownloadCSVFilters
from anyscale.cloud.models import (
    AWSConfig,
    Cloud,
    CloudPermissionLevel,
    CloudProvider,
    CloudResource,
    ComputeStack,
    CreateCloudCollaborator,
    FileStorage,
    GCPConfig,
    KubernetesConfig,
    NetworkingMode,
    NFSMountTarget,
    ObjectStorage,
)
from anyscale.commands import (
    aggregated_instance_usage_commands,
    cloud_commands,
    compute_config_commands,
    image_commands,
    job_commands,
    job_queue_commands,
    login_commands,
    logs_commands,
    machine_commands,
    machine_pool_commands,
    organization_invitation_commands,
    policy_commands,
    project_commands,
    resource_quota_commands,
    schedule_commands,
    scim_commands,
    service_account_commands,
    service_commands,
    user_commands,
    user_group_commands,
    workspace_commands_v2,
)
from anyscale.compute_config.models import (
    CloudDeployment as CloudDeploymentSelector,
    ComputeConfig,
    ComputeConfigListResult,
    ComputeConfigVersion,
    HeadNodeConfig,
    MarketType,
    MultiResourceComputeConfig,
    PhysicalResources,
    WorkerNodeGroupConfig,
)
from anyscale.image.models import ImageBuild, ImageBuildStatus
from anyscale.job.models import (
    JobConfig,
    JobLogMode,
    JobQueueConfig,
    JobQueueExecutionMode,
    JobQueueSpec,
    JobRunState,
    JobRunStatus,
    JobState,
    JobStatus,
)
from anyscale.job_queue.models import (
    ClusterState,
    ExecutionMode,
    JobQueueSortDirective,
    JobQueueSortField,
    JobQueueState,
    JobQueueStatus,
    SessionState,
    SortOrder,
)
from anyscale.organization_invitation.models import OrganizationInvitation
from anyscale.policy.models import (
    Policy,
    PolicyBinding,
    PolicyConfig,
    PolicySyncStatus,
    ResourcePolicy,
)
from anyscale.project.models import (
    CreateProjectCollaborator,
    Project,
    ProjectPermissionLevel,
    ProjectSortField,
    ProjectSortOrder,
)
from anyscale.resource_quota.models import CreateResourceQuota, Quota, ResourceQuota
from anyscale.schedule.models import ScheduleConfig, ScheduleState, ScheduleStatus
from anyscale.service.models import (
    RayGCSExternalStorageConfig,
    ServiceConfig,
    ServiceSortField,
    ServiceSortOrder,
    ServiceState,
    ServiceStatus,
    ServiceVersionState,
    ServiceVersionStatus,
    TracingConfig,
)
from anyscale.service_account.models import OrganizationPermissionLevel, ServiceAccount
from anyscale.user.models import (
    AdminCreatedUser,
    AdminCreateUser,
    AdminCreateUsers,
    User,
)
from anyscale.user_group.models import UserGroup
from anyscale.workspace.models import (
    Workspace,
    WorkspaceConfig,
    WorkspaceSortField,
    WorkspaceSortOrder,
    WorkspaceState,
)


# Defines all modules to be documented.
ALL_MODULES = [
    Module(
        title="Aggregated Instance Usage",
        filename="aggregated-instance-usage.md",
        cli_prefix="anyscale aggregated-instance-usage",
        cli_commands=[aggregated_instance_usage_commands.download_csv],
        sdk_prefix="anyscale.aggregated_instance_usage",
        sdk_commands=[anyscale.aggregated_instance_usage.download_csv,],
        models=[DownloadCSVFilters],
    ),
    Module(
        title="User",
        filename="user.md",
        cli_prefix="anyscale user",
        cli_commands=[
            user_commands.admin_batch_create,
            user_commands.list_users,
            user_commands.get_user,
            user_commands.list_permissions,
        ],
        sdk_prefix="anyscale.user",
        sdk_commands=[
            anyscale.user.admin_batch_create,
            anyscale.user.list,
            anyscale.user.get,
        ],
        models=[AdminCreateUser, AdminCreateUsers, AdminCreatedUser, User],
    ),
    Module(
        title="Project",
        filename="project-api.md",
        cli_prefix="anyscale project",
        cli_commands=[
            project_commands.get,
            project_commands.list,
            project_commands.create,
            project_commands.delete,
            project_commands.get_default,
            project_commands.add_collaborators,
        ],
        sdk_prefix="anyscale.project",
        sdk_commands=[
            anyscale.project.get,
            anyscale.project.list,
            anyscale.project.create,
            anyscale.project.delete,
            anyscale.project.get_default,
            anyscale.project.add_collaborators,
        ],
        models=[
            Project,
            ProjectSortField,
            ProjectSortOrder,
            CreateProjectCollaborator,
            ProjectPermissionLevel,
        ],
        legacy_cli_commands=[],
        legacy_sdk_commands={},
        legacy_sdk_models=[],
    ),
    Module(
        title="Job",
        filename="job-api.md",
        cli_prefix="anyscale job",
        cli_commands=[
            job_commands.submit,
            job_commands.status,
            job_commands.terminate,
            job_commands.archive,
            job_commands.delete,
            job_commands.logs,
            job_commands.wait,
            job_commands.list,
            job_commands.add_tags,
            job_commands.remove_tags,
            job_commands.list_tags,
        ],
        sdk_prefix="anyscale.job",
        sdk_commands=[
            anyscale.job.submit,
            anyscale.job.status,
            anyscale.job.terminate,
            anyscale.job.archive,
            anyscale.job.delete,
            anyscale.job.get_logs,
            anyscale.job.wait,
            anyscale.job.list,
            anyscale.job.add_tags,
            anyscale.job.remove_tags,
            anyscale.job.list_tags,
        ],
        models=[
            JobQueueSpec,
            JobQueueConfig,
            JobQueueExecutionMode,
            JobConfig,
            JobState,
            JobStatus,
            JobRunStatus,
            JobRunState,
            JobLogMode,
        ],
        cli_command_group_prefix={
            job_commands.add_tags: "tags",
            job_commands.remove_tags: "tags",
            job_commands.list_tags: "tags",
        },
        # The following commands are legacy
        legacy_sdk_commands={
            "create_job": anyscale.job.submit,
            "get_production_job": anyscale.job.status,
            "terminate_job": anyscale.job.terminate,
            "fetch_job_logs": anyscale.job.get_logs,
            "fetch_production_job_logs": anyscale.job.get_logs,
            "get_job_logs_download": anyscale.job.get_logs,
            "get_job_logs_stream": anyscale.job.get_logs,
            # limited support, no replacement yet
            "search_jobs": None,
            "list_production_jobs": None,
        },
        legacy_sdk_models=[
            "BaseJobStatus",
            "CreateClusterComputeConfig",
            "CreateJobQueueConfig",
            "CreateProductionJob",
            "CreateProductionJobConfig",
            "HaJobGoalStates",
            "HaJobStates",
            "Job",
            "JobListResponse",
            "JobQueueConfig",
            "JobQueueExecutionMode",
            "JobQueueSpec",
            "JobRunType",
            "JobStatus",
            "JobsSortField",
            "ProductionJob",
            "ProductionJobConfig",
            "ProductionJobStateTransition",
            "ProductionjobListResponse",
            "ProductionjobResponse",
            "RayRuntimeEnvConfig",
            "SortByClauseJobsSortField",
        ],
    ),
    Module(
        title="Schedule",
        filename="schedule-api.md",
        cli_prefix="anyscale schedule",
        cli_commands=[
            schedule_commands.apply,
            schedule_commands.list,
            schedule_commands.pause,
            schedule_commands.resume,
            schedule_commands.status,
            schedule_commands.trigger,
            schedule_commands.url,
            schedule_commands.delete,
        ],
        sdk_prefix="anyscale.schedule",
        sdk_commands=[
            anyscale.schedule.apply,
            anyscale.schedule.set_state,
            anyscale.schedule.status,
            anyscale.schedule.trigger,
            anyscale.schedule.list,
            anyscale.schedule.url,
            anyscale.schedule.delete,
        ],
        models=[ScheduleConfig, ScheduleState, ScheduleStatus],
        # The following commands are legacy
        legacy_sdk_commands={
            "create_or_update_schedule": anyscale.schedule.apply,
            "get_schedule": anyscale.schedule.status,
            "pause_schedule": anyscale.schedule.set_state,
            "run_schedule": anyscale.schedule.trigger,
            # limited support, no replacement yet
            "list_schedules": None,
        },
    ),
    Module(
        title="Service",
        filename="service-api.md",
        cli_prefix="anyscale service",
        cli_commands=[
            service_commands.list,
            service_commands.deploy,
            service_commands.status,
            service_commands.wait,
            service_commands.rollback,
            service_commands.terminate,
            service_commands.archive,
            service_commands.delete,
            service_commands.add_tags,
            service_commands.remove_tags,
            service_commands.list_tags,
        ],
        sdk_prefix="anyscale.service",
        sdk_commands=[
            anyscale.service.list,
            anyscale.service.deploy,
            anyscale.service.status,
            anyscale.service.wait,
            anyscale.service.rollback,
            anyscale.service.terminate,
            anyscale.service.archive,
            anyscale.service.delete,
            anyscale.service.add_tags,
            anyscale.service.remove_tags,
            anyscale.service.list_tags,
        ],
        models=[
            ServiceConfig,
            RayGCSExternalStorageConfig,
            TracingConfig,
            ServiceStatus,
            ServiceState,
            ServiceVersionStatus,
            ServiceVersionState,
            ServiceSortField,
            ServiceSortOrder,
        ],
        cli_command_group_prefix={
            service_commands.add_tags: "tags",
            service_commands.remove_tags: "tags",
            service_commands.list_tags: "tags",
        },
        # The following commands are legacy
        legacy_sdk_commands={
            "get_service": anyscale.service.status,
            "rollback_service": anyscale.service.rollback,
            "rollout_service": anyscale.service.deploy,
            "terminate_service": anyscale.service.terminate,
            # limited support, no replacement yet
            "list_services": None,
        },
        legacy_sdk_models=[
            "ApplyProductionServiceV2Model",
            "ApplyServiceModel",
            "ListServiceModel",
            "ListservicemodelListResponse",
            "ProductionServiceV2Model",
            "ProductionServiceV2VersionModel",
            "Productionservicev2ModelResponse",
            "RollbackServiceModel",
            "RolloutStrategy",
            "ServiceConfig",
            "ServiceEventCurrentState",
            "ServiceGoalStates",
            "ServiceModel",
            "ServiceObservabilityUrls",
            "ServiceSortField",
            "ServiceType",
            "ServiceVersionState",
            "ServicemodelListResponse",
            "ServicemodelResponse",
        ],
    ),
    Module(
        title="Compute Config",
        filename="compute-config-api.md",
        cli_prefix="anyscale compute-config",
        cli_commands=[
            compute_config_commands.create_compute_config,
            compute_config_commands.get_compute_config,
            compute_config_commands.archive_compute_config,
            compute_config_commands.list_compute_configs,
        ],
        sdk_prefix="anyscale.compute_config",
        sdk_commands=[
            anyscale.compute_config.create,
            anyscale.compute_config.get,
            anyscale.compute_config.get_default,
            anyscale.compute_config.archive,
            anyscale.compute_config.list,
        ],
        models=[
            ComputeConfig,
            ComputeConfigListResult,
            MultiResourceComputeConfig,
            HeadNodeConfig,
            WorkerNodeGroupConfig,
            MarketType,
            CloudDeploymentSelector,
            ComputeConfigVersion,
            PhysicalResources,
        ],
        legacy_sdk_commands={},
        legacy_sdk_models=[],
    ),
    Module(
        title="Service Account",
        filename="service-account.md",
        cli_prefix="anyscale service-account",
        cli_commands=[
            service_account_commands.create,
            service_account_commands.create_api_key,
            service_account_commands.list_service_accounts,
            service_account_commands.delete,
        ],
        sdk_prefix="anyscale.service_account",
        sdk_commands=[
            anyscale.service_account.create,
            anyscale.service_account.create_api_key,
            anyscale.service_account.list,
            anyscale.service_account.delete,
        ],
        models=[ServiceAccount, OrganizationPermissionLevel],
    ),
    Module(
        title="Image",
        filename="image.md",
        cli_prefix="anyscale image",
        cli_commands=[
            image_commands.build,
            image_commands.get,
            image_commands.list,
            image_commands.register,
            image_commands.archive,
        ],
        sdk_prefix="anyscale.image",
        sdk_commands=[
            anyscale.image.build,
            anyscale.image.get,
            anyscale.image.list,
            anyscale.image.register,
            anyscale.image.archive,
        ],
        models=[ImageBuildStatus, ImageBuild],
    ),
    Module(
        title="Cloud",
        filename="cloud.md",
        cli_prefix="anyscale cloud",
        cli_commands=[
            cloud_commands.setup_cloud,
            cloud_commands.register_cloud,
            cloud_commands.cloud_update,
            cloud_commands.cloud_delete,
            cloud_commands.cloud_verify,
            cloud_commands.list_cloud,
            cloud_commands.cloud_resource_create,
            cloud_commands.cloud_resource_delete,
            cloud_commands.cloud_config_get,
            cloud_commands.cloud_config_update,
            cloud_commands.cloud_set_default,
            cloud_commands.add_collaborators,
            cloud_commands.get_cloud,
            cloud_commands.get_default_cloud,
            cloud_commands.terminate_system_cluster,
        ],
        sdk_prefix="anyscale.cloud",
        sdk_commands=[
            anyscale.cloud.add_collaborators,
            anyscale.cloud.list,
            anyscale.cloud.get,
            anyscale.cloud.get_default,
            anyscale.cloud.terminate_system_cluster,
        ],
        models=[
            Cloud,
            CloudPermissionLevel,
            CreateCloudCollaborator,
            CloudResource,
            ComputeStack,
            CloudProvider,
            NetworkingMode,
            ObjectStorage,
            FileStorage,
            NFSMountTarget,
            AWSConfig,
            GCPConfig,
            KubernetesConfig,
        ],
        cli_command_group_prefix={
            cloud_commands.cloud_resource_create: "resource",
            cloud_commands.cloud_resource_delete: "resource",
            cloud_commands.cloud_config_get: "config",
            cloud_commands.cloud_config_update: "config",
        },
        legacy_sdk_commands={
            # limited support, no replacement yet
            "get_cloud": None,
            "get_default_cloud": None,
            "search_clouds": None,
        },
        legacy_sdk_models=[
            "Cloud",
            "CloudConfig",
            "CloudListResponse",
            "CloudProviders",
            "CloudResponse",
            "CloudState",
            "CloudStatus",
            "CloudTypes",
            "CloudVersion",
            "CloudsQuery",
        ],
    ),
    Module(
        title="Logs",
        filename="logs.md",
        cli_prefix="anyscale logs",
        cli_commands=[
            logs_commands.anyscale_logs_workspace,
            logs_commands.anyscale_logs_service,
            logs_commands.anyscale_logs_job,
            logs_commands.anyscale_logs_cluster,
        ],
        sdk_prefix="anyscale.logs",
        sdk_commands=[],
        models=[],
    ),
    Module(
        title="Workspaces",
        filename="workspaces.md",
        cli_prefix="anyscale workspace_v2",
        cli_commands=[
            workspace_commands_v2.create,
            workspace_commands_v2.start,
            workspace_commands_v2.terminate,
            workspace_commands_v2.status,
            workspace_commands_v2.wait,
            workspace_commands_v2.ssh,
            workspace_commands_v2.run_command,
            workspace_commands_v2.pull,
            workspace_commands_v2.push,
            workspace_commands_v2.update,
            workspace_commands_v2.get,
            workspace_commands_v2.list,
            workspace_commands_v2.add_tags,
            workspace_commands_v2.remove_tags,
            workspace_commands_v2.list_tags,
        ],
        sdk_prefix="anyscale.workspace",
        sdk_commands=[
            anyscale.workspace.create,
            anyscale.workspace.start,
            anyscale.workspace.terminate,
            anyscale.workspace.status,
            anyscale.workspace.wait,
            anyscale.workspace.generate_ssh_config_file,
            anyscale.workspace.run_command,
            anyscale.workspace.list,
            anyscale.workspace.add_tags,
            anyscale.workspace.remove_tags,
            anyscale.workspace.list_tags,
        ],
        models=[
            WorkspaceConfig,
            WorkspaceState,
            Workspace,
            WorkspaceSortField,
            WorkspaceSortOrder,
        ],
        cli_command_group_prefix={
            workspace_commands_v2.add_tags: "tags",
            workspace_commands_v2.remove_tags: "tags",
            workspace_commands_v2.list_tags: "tags",
        },
    ),
    Module(
        title="Job Queue",
        filename="job-queue.md",
        cli_prefix="anyscale job-queue",
        cli_commands=[
            job_queue_commands.list_job_queues,
            job_queue_commands.update_job_queue,
            job_queue_commands.status,
            job_queue_commands.archive_job_queue,
            job_queue_commands.terminate_job_queue,
            job_queue_commands.delete_job_queue,
            job_queue_commands.add_tags,
            job_queue_commands.remove_tags,
            job_queue_commands.list_tags,
        ],
        sdk_prefix="anyscale.job_queue",
        sdk_commands=[
            anyscale.job_queue.list,
            anyscale.job_queue.status,
            anyscale.job_queue.update,
            anyscale.job_queue.archive,
            anyscale.job_queue.terminate,
            anyscale.job_queue.delete,
            anyscale.job_queue.add_tags,
            anyscale.job_queue.remove_tags,
            anyscale.job_queue.list_tags,
        ],
        models=[
            JobQueueStatus,
            JobQueueState,
            ExecutionMode,
            ClusterState,
            JobQueueSortField,
            JobQueueSortDirective,
            SortOrder,
            SessionState,
        ],
        cli_command_group_prefix={
            job_queue_commands.add_tags: "tags",
            job_queue_commands.remove_tags: "tags",
            job_queue_commands.list_tags: "tags",
        },
    ),
    Module(
        title="Machine Pool",
        filename="machine-pool.md",
        cli_prefix="anyscale machine-pool",
        cli_commands=[
            machine_pool_commands.create_machine_pool,
            machine_pool_commands.update_machine_pool,
            machine_pool_commands.delete_machine_pool,
            machine_pool_commands.attach_machine_pool_to_cloud,
            machine_pool_commands.detach_machine_pool_from_cloud,
            machine_pool_commands.list_machine_pools,
            machine_pool_commands.describe,
        ],
        sdk_prefix="anyscale.machine_pool",
        sdk_commands=[],
        models=[],
    ),
    Module(
        title="Cluster",
        filename="cluster.md",
        cli_prefix="anyscale cluster",
        cli_commands=[],
        sdk_prefix="anyscale.cluster",  # fake name used as id, cluster has no SDK commands
        sdk_commands=[],
        models=[],
        legacy_sdk_commands={
            "launch_cluster": None,
            "launch_cluster_with_new_cluster_environment": None,
            "create_cluster": None,
            "delete_cluster": None,
            "get_cluster": None,
            "search_clusters": None,
            "start_cluster": None,
            "terminate_cluster": None,
            "update_cluster": None,
        },
        legacy_sdk_models=[
            "ClusterHeadNodeInfo",
            "ClusterListResponse",
            "ClusterManagementStackVersions",
            "ClusterOperation",
            "ClusterResponse",
            "ClusterServicesUrls",
            "ClusterState",
            "ClusterStatus",
            "ClusterStatusDetails",
            "ClusteroperationResponse",
            "ClustersQuery",
            "CreateCluster",
            "StartClusterOptions",
            "TerminateClusterOptions",
            "UpdateCluster",
        ],
    ),
    Module(
        title="Resource quotas",
        filename="resource-quotas.md",
        cli_prefix="anyscale resource-quota",
        cli_commands=[
            resource_quota_commands.create,
            resource_quota_commands.list_resource_quotas,
            resource_quota_commands.delete,
            resource_quota_commands.enable,
            resource_quota_commands.disable,
        ],
        sdk_prefix="anyscale.resource_quota",
        sdk_commands=[
            anyscale.resource_quota.create,
            anyscale.resource_quota.list,
            anyscale.resource_quota.delete,
            anyscale.resource_quota.enable,
            anyscale.resource_quota.disable,
        ],
        models=[CreateResourceQuota, Quota, ResourceQuota],
    ),
    Module(
        title="Organization Invitation",
        filename="organization-invitation.md",
        cli_prefix="anyscale organization-invitation",
        cli_commands=[
            organization_invitation_commands.create,
            organization_invitation_commands.list,
            organization_invitation_commands.delete,
        ],
        sdk_prefix="anyscale.organization-invitation",
        sdk_commands=[
            anyscale.organization_invitation.create,
            anyscale.organization_invitation.list,
            anyscale.organization_invitation.delete,
        ],
        models=[OrganizationInvitation],
    ),
    Module(
        title="User Group",
        filename="user-group.md",
        cli_prefix="anyscale user-group",
        cli_commands=[
            user_group_commands.list_user_groups,
            user_group_commands.get_user_group,
            user_group_commands.list_memberships,
        ],
        sdk_prefix="anyscale.user_group",
        sdk_commands=[anyscale.user_group.list, anyscale.user_group.get,],
        models=[UserGroup],
    ),
    Module(
        title="Policy",
        filename="policy.md",
        cli_prefix="anyscale policy",
        cli_commands=[
            policy_commands.set_policy,
            policy_commands.get_policy,
            policy_commands.list_policies,
        ],
        sdk_prefix="anyscale.policy",
        sdk_commands=[anyscale.policy.set, anyscale.policy.get, anyscale.policy.list,],
        models=[Policy, PolicyBinding, PolicyConfig, PolicySyncStatus, ResourcePolicy],
    ),
    Module(
        title="SCIM",
        filename="scim.md",
        cli_prefix="anyscale scim",
        cli_commands=[scim_commands.enforce_group_permissions,],
        sdk_prefix="anyscale.scim",
        sdk_commands=[],
        models=[],
    ),
    Module(
        title="Other",
        filename="other.md",
        cli_prefix="anyscale",
        cli_commands=[
            login_commands.anyscale_login,
            login_commands.anyscale_logout,
            scripts.version_cli,
        ],
        sdk_prefix="",
        sdk_commands=[],
        models=[],
        # The following commands are legacy
        legacy_cli_prefix="anyscale",
        legacy_cli_commands=[
            machine_commands.delete_machine,
            machine_commands.list_machines,
        ],
        legacy_sdk_commands={
            "partial_update_organization": None,
            "upsert_sso_config": None,
            "upsert_test_sso_config": None,
        },
        legacy_sdk_models=[
            "AccessConfig",
            "ArchiveStatus",
            "BASEIMAGESENUM",
            "BaseJobStatus",
            "Build",
            "BuildResponse",
            "BuildStatus",
            "CreateSSOConfig",
            "GrpcProtocolConfig",
            "HTTPValidationError",
            "HttpProtocolConfig",
            "IdleTerminationStatus",
            "ListResponseMetadata",
            "NodeType",
            "OperationError",
            "OperationProgress",
            "OperationResult",
            "Organization",
            "OrganizationResponse",
            "PageQuery",
            "Protocols",
            "PythonModules",
            "PythonVersion",
            "RayGCSExternalStorageConfig",
            "Resources",
            "SSOConfig",
            "SSOMode",
            "SUPPORTEDBASEIMAGESENUM",
            "SortOrder",
            "SsoconfigResponse",
            "StaticSSOConfig",
            "TextQuery",
            "TracingConfig",
            "UpdateOrganization",
            "UserServiceAccessTypes",
            "WorkerNodeType",
        ],
    ),
]


@click.command(help="Generate markdown docs for the Anyscale CLI & SDK.")
@click.argument("output_dir")
@click.option(
    "-r",
    "--remove-existing",
    is_flag=True,
    default=False,
    help="If set, all files in the 'output_dir' that were not generated will be removed.",
)
def generate(
    output_dir: str, *, remove_existing: bool = False,
):
    if not os.path.isdir(output_dir):
        raise RuntimeError(f"output_dir '{output_dir}' does not exist.")

    gen = MarkdownGenerator(ALL_MODULES)

    generated_files = set()
    os.makedirs(output_dir, exist_ok=True)

    # Create legacy subdirectory
    legacy_dir = os.path.join(output_dir, "legacy")
    os.makedirs(legacy_dir, exist_ok=True)

    for filename, content in gen.generate().items():
        generated_files.add(filename)
        full_path = os.path.join(output_dir, filename)

        # Create directory if it doesn't exist (for legacy/ files)
        dir_path = os.path.dirname(full_path)
        if dir_path:
            os.makedirs(dir_path, exist_ok=True)

        print(f"Writing output file {full_path}")
        with open(full_path, "w") as f:
            f.write(content)

    if remove_existing:
        # Get all existing files (including in subdirectories)
        existing_files = set()
        for root, _dirs, files in os.walk(output_dir):
            for file in files:
                rel_path = os.path.relpath(os.path.join(root, file), output_dir)
                existing_files.add(rel_path)

        # Remove files that weren't generated
        to_remove = existing_files - generated_files
        for path in to_remove:
            full_path = os.path.join(output_dir, path)
            if os.path.exists(full_path):
                print(f"Removing existing file {full_path}")
                os.unlink(full_path)


if __name__ == "__main__":
    generate()



from contextlib import contextmanager
from copy import deepcopy
import datetime
from enum import Enum
import ipaddress
import json
import logging
import os
import random
import shutil
import string
import sys
import time
from typing import (
    Any,
    Callable,
    cast,
    Dict,
    Generator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)
from urllib.parse import urljoin

import boto3
from boto3.resources.base import ServiceResource as Boto3Resource
from botocore.config import Config
from botocore.exceptions import ClientError, NoRegionError
import click
from packaging import version
import requests
from requests import Response

from anyscale.authenticate import get_auth_api_client
from anyscale.aws_iam_policies import ANYSCALE_IAM_POLICIES, AnyscaleIAMPolicy
from anyscale.cli_logger import BlockLogger, CloudSetupLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi as ProductApi
from anyscale.client.openapi_client.models import (
    AWSMemoryDBClusterConfig,
    ServiceEventCurrentState,
)
from anyscale.client.openapi_client.models.cloud_analytics_event_cloud_resource import (
    CloudAnalyticsEventCloudResource,
)
from anyscale.client.openapi_client.models.decorated_compute_template import (
    DecoratedComputeTemplate,
)
from anyscale.client.openapi_client.models.ha_job_states import HaJobStates
from anyscale.client.openapi_client.models.user_info import UserInfo
from anyscale.cluster_compute import get_cluster_compute_from_name
import anyscale.conf
from anyscale.conf import MINIMUM_RAY_VERSION
from anyscale.feature_flags import FLAG_DEFAULT_WORKING_DIR_FOR_PROJ
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as BaseApi
from anyscale.sdk.anyscale_client.models.cluster import Cluster
from anyscale.sdk.anyscale_client.models.compute_template import ComputeTemplate
import anyscale.shared_anyscale_utils.conf as shared_anyscale_conf
from anyscale.utils.cloud_utils import CloudSetupError


logger = logging.getLogger(__file__)

BOTO_MAX_RETRIES = 5
PROJECT_NAME_ENV_VAR = "ANYSCALE_PROJECT_NAME"

VALID_BYOD_PYTHON_VERSIONS = ["py36", "py37", "py38", "py39"]

REDIS_TLS_ADDRESS_PREFIX = "rediss://"

log = BlockLogger()  # Anyscale CLI Logger

VPC_CIDR_RANGE = "10.0.0.0/16"

DEFAULT_RAY_VERSION = "1.7.0"

# The V2 stack has some delay before changing the state of a new cluster from
# Terminated to StartingUp. Allow the terminated state for the duration of this
# grace period before erroring.
TERMINATED_STATE_GRACE_PERIOD_SECONDS = 5 * 60  # 5 minutes

MEMORY_DB_OUTPUT = """  MemoryDB:
    Description: MemoryDB cluster
    Value:
      Fn::ToJsonString:
        arn: !GetAtt MemoryDB.ARN
        ClusterEndpointAddress: !GetAtt MemoryDB.ClusterEndpoint.Address"""

MEMORY_DB_RESOURCE = """  MemoryDBSubnetGroup:
    Type: AWS::MemoryDB::SubnetGroup
    Properties:
      Description: Anyscale managed MemoryDB subnet group
      SubnetGroupName: !Ref AWS::StackName
      SubnetIds:
{}
      Tags:
        - Key: anyscale-cloud-id
          Value: !Ref CloudID

  MemoryDBParameterGroup:
    Type: AWS::MemoryDB::ParameterGroup
    Properties:
      Description: Parameter group for anyscale managed MemoryDB
      Family: memorydb_redis7
      ParameterGroupName:  !Ref AWS::StackName
      Tags:
        - Key: anyscale-cloud-id
          Value: !Ref CloudID

  MemoryDB:
    Type: AWS::MemoryDB::Cluster
    Properties:
      ACLName: open-access
      Description: Anyscale managed MemoryDB
      ClusterName: !Ref AWS::StackName
      NodeType: db.t4g.small
      Port: !Ref MemoryDBRedisPort
      SubnetGroupName: !Ref MemoryDBSubnetGroup
      SecurityGroupIds:
        - !Ref AnyscaleSecurityGroup
      EngineVersion: "7.0"
      ParameterGroupName: !Ref MemoryDBParameterGroup
      TLSEnabled: true
      Tags:
        - Key: anyscale-cloud-id
          Value: !Ref CloudID"""

# Some resources (e.g. memorystore) take a long time to create, so we need to increase the timeout.
GCP_DEPLOYMENT_MANAGER_TIMEOUT_SECONDS_LONG = 600  # 10 minutes


class AnyscaleJSONEncoder(json.JSONEncoder):
    """Custom JSON encoder for Anyscale models, handling datetime objects."""

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


def confirm(msg: str, yes: bool) -> Optional[bool]:
    return None if yes else click.confirm(msg, abort=True)


class SharedStorageType(str, Enum):
    """Enum for shared storage types."""

    OBJECT_STORAGE = "object-storage"
    NFS = "nfs"


class AnyscaleEndpointFormatter:
    def __init__(self, host: Optional[str] = None):
        self.host = host or shared_anyscale_conf.ANYSCALE_HOST

    def get_endpoint(self, endpoint: str) -> str:
        return str(urljoin(self.host, endpoint))

    def get_job_endpoint(self, job_id: str) -> str:
        return self.get_endpoint(f"/jobs/{job_id}")

    def get_schedule_endpoint(self, schedule_id: str) -> str:
        return self.get_endpoint(f"/scheduled-jobs/{schedule_id}")


def get_endpoint(endpoint: str, host: Optional[str] = None) -> str:
    return str(urljoin(host or shared_anyscale_conf.ANYSCALE_HOST, endpoint))


def send_json_request_raw(
    endpoint: str,
    json_args: Dict[str, Any],
    method: str = "GET",
    cli_token: Optional[str] = None,
    host: Optional[str] = None,
) -> Response:
    get_auth_api_client(cli_token=cli_token, host=host)

    url = get_endpoint(endpoint, host=host)
    cookies = {"cli_token": cli_token or anyscale.conf.CLI_TOKEN or ""}
    try:
        if method == "GET":
            resp = requests.get(url, params=json_args, cookies=cookies)
        elif method == "POST":
            resp = requests.post(url, json=json_args, cookies=cookies)
        elif method == "DELETE":
            resp = requests.delete(url, json=json_args, cookies=cookies)
        elif method == "PATCH":
            resp = requests.patch(url, data=json_args, cookies=cookies)
        elif method == "PUT":
            resp = requests.put(url, json=json_args, cookies=cookies)
        else:
            raise AssertionError(f"unknown method {method}")
    except requests.exceptions.ConnectionError:
        raise click.ClickException(f"Failed to connect to anyscale server at {url}")

    return resp


def send_json_request(
    endpoint: str,
    json_args: Dict[str, Any],
    method: str = "GET",
    cli_token: Optional[str] = None,
    host: Optional[str] = None,
) -> Dict[str, Any]:
    resp = send_json_request_raw(
        endpoint, json_args, method=method, cli_token=cli_token, host=host,
    )

    if not resp.ok:
        if resp.status_code == 500:
            raise click.ClickException(
                "There was an internal error in this command. "
                "Please report this to the Anyscale team at support@anyscale.com "
                "with the token '{}'.".format(resp.headers["x-trace-id"])
            )

        raise click.ClickException(f"{resp.status_code}: {resp.text}.")

    if resp.status_code == 204:
        return {}

    json_resp: Dict[str, Any] = resp.json()
    if "error" in json_resp:
        raise click.ClickException("{}".format(json_resp["error"]))

    return json_resp


def deserialize_datetime(s: str) -> datetime.datetime:
    if sys.version_info < (3, 7) and s[-3:-2] == ":":
        s = s[:-3] + s[-2:]

    return datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%f%z")


def humanize_timestamp(timestamp: datetime.datetime) -> str:
    delta = datetime.datetime.now(datetime.timezone.utc) - timestamp
    offset = float(delta.seconds + (delta.days * 60 * 60 * 24))
    delta_s = int(offset % 60)
    offset /= 60
    delta_m = int(offset % 60)
    offset /= 60
    delta_h = int(offset % 24)
    offset /= 24
    delta_d = int(offset)

    if delta_d >= 1:
        return "{} day{} ago".format(delta_d, "s" if delta_d > 1 else "")
    if delta_h > 0:
        return "{} hour{} ago".format(delta_h, "s" if delta_h > 1 else "")
    if delta_m > 0:
        return "{} minute{} ago".format(delta_m, "s" if delta_m > 1 else "")
    else:
        return "{} second{} ago".format(delta_s, "s" if delta_s > 1 else "")


def get_requirements(requirements_path: str) -> str:
    with open(requirements_path) as f:
        return f.read()


def _resource(name: str, region: str) -> Boto3Resource:
    boto_config = Config(retries={"max_attempts": BOTO_MAX_RETRIES})
    return boto3.resource(name, region, config=boto_config)  # type: ignore


def _client(name: str, region: str) -> Any:
    return _resource(name, region).meta.client


def _get_role(
    role_name: str, region: str, boto3_session: Optional[boto3.Session] = None
) -> Optional[Boto3Resource]:
    iam = (
        _resource("iam", region)
        if boto3_session is None
        else boto3_session.resource("iam")
    )
    role = iam.Role(role_name)  # type: ignore
    try:
        role.load()
        return role  # type: ignore
    except ClientError as exc:
        if exc.response.get("Error", {}).get("Code") == "NoSuchEntity":
            return None
        else:
            raise exc


def _get_aws_efs_mount_target_ip(boto3_session: Any, efs_id: str) -> Optional[str]:
    client = boto3_session.client("efs")
    try:
        mount_targets_response = client.describe_mount_targets(FileSystemId=efs_id)
        if not mount_targets_response.get("MountTargets"):
            logger.warning(f"EFS with id {efs_id} does not contain mount targets.")
            return None
        return mount_targets_response.get("MountTargets")[0].get("IpAddress")
    except ClientError as e:
        if e.response["Error"]["Code"] == "FileSystemNotFound":
            return None
        raise e


def _get_subnet(
    subnet_arn: str, region: str, logger: CloudSetupLogger
) -> Optional[Boto3Resource]:
    ec2 = _resource("ec2", region)  # TODO: take a resource as an argument
    subnet = ec2.Subnet(subnet_arn)  # type: ignore
    try:
        subnet.load()
        return subnet  # type: ignore
    except ClientError as e:
        if e.response["Error"]["Code"] == "InvalidSubnetID.NotFound":
            logger.log_resource_error(
                CloudAnalyticsEventCloudResource.AWS_SUBNET,
                CloudSetupError.RESOURCE_NOT_FOUND,
            )
            raise click.ClickException(
                f"{subnet_arn} does not exist. Please make sure the subnet arn is correct and the subnet is in the same region as the cloud."
            )
        else:
            logger.log_resource_exception(
                CloudAnalyticsEventCloudResource.AWS_SUBNET, e
            )
        raise e


def _get_memorydb_cluster_config(
    memorydb_cluster_id: str, region: str, logger: CloudSetupLogger
) -> Optional[AWSMemoryDBClusterConfig]:
    try:
        memorydb_client = boto3.client("memorydb", region_name=region)
        response = memorydb_client.describe_clusters(ClusterName=memorydb_cluster_id)

        if not response.get("Clusters") or not response.get("Clusters")[0]:
            logger.log_resource_error(
                CloudAnalyticsEventCloudResource.AWS_MEMORYDB,
                CloudSetupError.RESOURCE_NOT_FOUND,
            )
            raise click.ClickException(
                f"MemoryDB cluster with id {memorydb_cluster_id} does not exist."
            )
        cluster = response["Clusters"][0]
        if cluster["Status"] != "available":
            logger.log_resource_error(
                CloudAnalyticsEventCloudResource.AWS_MEMORYDB,
                CloudSetupError.MEMORYDB_CLUSTER_UNAVAILABLE,
            )
            raise click.ClickException(
                f"MemoryDB cluster with id {memorydb_cluster_id} is not currently available. Please make sure the cluster is available and try again."
            )
        endpoint = cluster["ClusterEndpoint"]

        return AWSMemoryDBClusterConfig(
            id=cluster["ARN"],
            endpoint=f"{REDIS_TLS_ADDRESS_PREFIX}{endpoint['Address']}:{endpoint['Port']}",
        )
    except ClientError as e:
        if e.response["Error"]["Code"] == "ClusterNotFoundFault":
            logger.log_resource_error(
                CloudAnalyticsEventCloudResource.AWS_MEMORYDB,
                CloudSetupError.RESOURCE_NOT_FOUND,
            )
        else:
            logger.log_resource_exception(
                CloudAnalyticsEventCloudResource.AWS_MEMORYDB, e
            )
        raise e


def get_available_regions(boto3_session: Optional[boto3.Session] = None) -> List[str]:
    if boto3_session is None:
        boto3_session = boto3.Session()
    try:
        client = boto3_session.client("ec2")
    except NoRegionError:
        # If there is no region, default to `us-west-2`
        client = boto3_session.client("ec2", region_name="us-west-2")
    return [region["RegionName"] for region in client.describe_regions()["Regions"]]


def get_availability_zones(
    region: str, boto3_session: Optional[boto3.Session] = None
) -> Dict[str, str]:
    """
    Returns a mapping of availability zone ids to names for the given region.
    """
    if boto3_session is None:
        boto3_session = boto3.Session()
    client = boto3_session.client("ec2", region_name=region)
    return {
        zone["ZoneId"]: zone["ZoneName"]
        for zone in client.describe_availability_zones()["AvailabilityZones"]
    }


def get_memorydb_supported_zones(
    region: str, zone_ids_to_names: Dict[str, str]
) -> List[str]:
    """
    Returns a list of supported zone names for MemoryDB in the given region.
    """
    with open(f"{anyscale.conf.ROOT_DIR_PATH}/memorydb_supported_zones.json") as f:
        # There's no API to get the supported zones for MemoryDB, so we need to read it from a file.
        memorydb_supported_zones = json.load(f)

    supported_zone_ids_for_cloud_region = memorydb_supported_zones.get(region, [])
    if len(supported_zone_ids_for_cloud_region) == 0:
        raise click.ClickException(f"MemoryDB is not supported in the region {region}.")
    return [
        zone_ids_to_names[zone_id] for zone_id in supported_zone_ids_for_cloud_region
    ]


def get_project_directory_name(project_id: str, api_client: ProductApi = None) -> str:
    if api_client is None:
        api_client = get_auth_api_client().api_client

    # TODO (yiran): return error early if project doesn't exist.
    resp = api_client.get_project_api_v2_projects_project_id_get(project_id)
    if api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get(
        FLAG_DEFAULT_WORKING_DIR_FOR_PROJ
    ).result.is_on:
        directory_name = resp.result.directory_name
    else:
        directory_name = resp.result.name
    assert len(directory_name) > 0, "Empty directory name found."
    return cast(str, directory_name)


def get_working_dir(project_id: str, api_client: ProductApi = None) -> str:
    project_directory_name = get_project_directory_name(project_id, api_client)
    return f"/home/ray/{project_directory_name}"


# Python 3.7 and before have different wheel name conventions.
# Though product today only support Python 3.9 or above.
_LEGACY_PY_VERSIONS = ["3" + str(x) for x in range(8)]


def get_wheel_url(
    ray_commit: str,
    ray_version: str,
    py_version: Optional[str] = None,
    sys_platform: Optional[str] = None,
) -> str:
    """Return S3 URL for the given release spec or 'latest'."""
    if py_version is None:
        py_version = "".join(str(x) for x in sys.version_info[0:2])
    if sys_platform is None:
        sys_platform = sys.platform

    if sys_platform == "darwin":
        if py_version not in _LEGACY_PY_VERSIONS:
            platform = "macosx_10_15_x86_64"
        else:
            platform = "macosx_10_15_intel"
    elif sys_platform == "win32":
        platform = "win_amd64"
    else:
        platform = "manylinux2014_x86_64"

    py_version_malloc = (
        f"{py_version}m" if py_version in _LEGACY_PY_VERSIONS else py_version
    )

    if "dev" in ray_version:
        ray_release = f"master/{ray_commit}"
    else:
        ray_release = f"releases/{ray_version}/{ray_commit}"
    return (
        "https://s3-us-west-2.amazonaws.com/ray-wheels/"
        "{}/ray-{}-cp{}-cp{}-{}.whl".format(
            ray_release, ray_version, py_version, py_version_malloc, platform
        )
    )


@contextmanager
def updating_printer() -> Generator[Callable[[str], None], None, None]:
    cols, _ = shutil.get_terminal_size()

    def print_status(status: str) -> None:
        lines = status.splitlines()
        first_line = lines[0]
        truncated_first_line = (
            first_line[0:cols]
            if len(first_line) <= cols and len(lines) == 1
            else (first_line[0 : cols - 3] + "...")
        )
        # Clear the line first
        print("\r" + " " * cols, end="\r")
        print(truncated_first_line, end="", flush=True)

    try:
        yield print_status
    finally:
        # Clear out the status and return to the beginning to reprint
        print("\r" + " " * cols, end="\r", flush=True)


def wait_for_session_start(  # noqa: PLR0912
    project_id: str,
    session_name: str,
    api_client: Optional[ProductApi] = None,
    log: BlockLogger = log,
    block_label: Optional[str] = None,
) -> str:
    if block_label:
        log.info(
            f"Waiting for cluster {BlockLogger.highlight(session_name)} to start. This may take a few minutes",
            block_label=block_label,
        )
    else:
        log.info(
            f"Waiting for cluster {session_name} to start. This may take a few minutes"
        )

    if api_client is None:
        api_client = get_auth_api_client().api_client

    start_time = time.time()
    with updating_printer() as print_status:
        while True:
            sessions = api_client.list_sessions_api_v2_sessions_get(
                project_id=project_id, name=session_name, active_only=False
            ).results

            if len(sessions) > 0:
                session = sessions[0]

                # TODO: Remove "session.host_name" check once https://github.com/anyscale/product/issues/15502 is fixed
                # A cluster may have "running" state while its dns is not set up yet. When DNS is not ready,
                # many cluster operations, like ray job submission will fail. So we wait until DNS info
                # is ready.
                if (
                    session.state == "Running"
                    and session.pending_state is None
                    and session.host_name
                ):
                    return cast(str, session.id)

                # Check for start up errors
                if (
                    session.state_data
                    and session.state_data.startup
                    and session.state_data.startup.startup_error
                ):
                    raise click.ClickException(
                        f"Error while starting cluster {session_name}: {session.state_data.startup.startup_error}"
                    )
                elif (
                    session.state
                    and "Errored" in session.state
                    and session.pending_state is None
                ):
                    raise click.ClickException(
                        f"Error while starting cluster {session_name}: Cluster startup failed due to an error ({session.state})."
                    )
                elif (
                    session.state
                    and session.state in {"Terminated", "Stopped"}
                    and session.pending_state is None
                ):
                    if time.time() - start_time < TERMINATED_STATE_GRACE_PERIOD_SECONDS:
                        # The V2 stack has some delay before transitioning from Terminated to StartingUp.
                        # Don't error until the grace period has passed
                        print_status("Waiting for start up...")
                    else:
                        # Cluster is created in Terminated state; Check pending state to see if it is pending transition.
                        raise click.ClickException(
                            f"Error while starting cluster {session_name}: Cluster is still in stopped/terminated state."
                        )
                elif (
                    session.state_data
                    and session.state_data.startup
                    and session.state_data.startup.startup_progress
                ):
                    # Print the latest status
                    print_status(
                        "Starting up " + session.state_data.startup.startup_progress
                    )
                elif (
                    session.state != "StartingUp"
                    and session.pending_state == "StartingUp"
                ):
                    print_status("Waiting for start up...")
            else:
                raise click.ClickException(
                    f"Error while starting cluster {session_name}: Cluster doesn't exist."
                )

            time.sleep(2)


def populate_session_args(cluster_config_str: str, config_file_name: str) -> str:
    import jinja2  # noqa: PLC0415 - codex_reason("gpt5.2", "optional templating dependency for config rendering")

    env = jinja2.Environment()
    t = env.parse(cluster_config_str)
    for elem in t.body[0].nodes:  # type: ignore
        if isinstance(elem, jinja2.nodes.Getattr) and elem.attr not in os.environ:
            prefixed_command = " ".join(
                [f"{elem.attr}=<value>", "anyscale"] + sys.argv[1:]
            )
            raise click.ClickException(
                f"\tThe environment variable {elem.attr} was not set, yet it is required "
                f"for configuration file {config_file_name}.\n\tPlease specify {elem.attr} "
                f"by prefixing the command.\n\t\t{prefixed_command}"
            )

    template = jinja2.Template(cluster_config_str)
    cluster_config_filled = template.render(env=os.environ)
    return cluster_config_filled


def get_user_info() -> Optional[UserInfo]:
    try:
        api_client = get_auth_api_client().api_client
    except click.exceptions.ClickException:
        return None
    return api_client.get_user_info_api_v2_userinfo_get().result


def generate_slug(length: int = 6) -> str:
    return "".join(random.choices(string.ascii_lowercase + string.digits, k=length))


def validate_non_negative_arg(ctx, param, value):  # noqa: ARG001
    """
    Checks that an integer option to click command is non-negative.
    """
    if value is not None and value < 0:
        raise click.ClickException(
            f"Please specify a non-negative value for {param.opts[0]}"
        )
    return value


def validate_service_state_filter(
    ctx, param, value: Tuple[str, ...]  # noqa: ARG001
) -> List[str]:
    """Validate ServiceEventCurrentState values."""
    if not value:
        return []

    allowable_values_upper = {
        s.upper() for s in ServiceEventCurrentState.allowable_values
    }
    allowed_values_str = ", ".join(ServiceEventCurrentState.allowable_values)

    for state_str in value:
        state_upper = state_str.upper()
        if state_upper not in allowable_values_upper:
            raise click.ClickException(
                f"'{state_str}' is not a valid value for {param.opts[0]}. Allowed values: {allowed_values_str}"
            )

    return [s.upper() for s in value]


def validate_workspace_state_filter(
    ctx, param, value: Tuple[str, ...]  # noqa: ARG001
) -> List[str]:
    """Validate WorkspaceState values (case-insensitive) and return uppercase."""
    if not value:
        return []

    # Import here to avoid circular dependency
    from anyscale.workspace.models import (  # noqa: PLC0415 - codex_reason("gpt5.2", "avoid circular dependency during CLI import")
        WorkspaceState,
    )

    allowable_values_upper = {s.value.upper() for s in WorkspaceState}
    allowed_values_str = ", ".join(s.value for s in WorkspaceState)

    for state_str in value:
        state_upper = state_str.upper()
        if state_upper not in allowable_values_upper:
            raise click.ClickException(
                f"'{state_str}' is not a valid value for {param.opts[0]}. Allowed values: {allowed_values_str}"
            )

    # Return uppercase WorkspaceState values (e.g., "RUNNING")
    return [s.upper() for s in value]


def _update_external_ids_for_policy(
    original_policy: Dict[str, Any], new_external_id: str
):
    """Gets All External IDs From policy Dict."""
    policy = deepcopy(original_policy)
    external_ids = [
        statement.setdefault("Condition", {})
        .setdefault("StringEquals", {})
        .setdefault("sts:ExternalId", [])
        for statement in policy.get("Statement", [])
    ]

    external_ids = [
        [i, new_external_id] if isinstance(i, str) else i + [new_external_id]
        for i in external_ids
    ]

    # remove duplicate external IDs
    external_ids = [sorted(set(ids)) for ids in external_ids]

    _ = [
        policy["Statement"][i]["Condition"]["StringEquals"].update(
            {"sts:ExternalId": external_ids[i]}
        )
        for i in range(len(policy.get("Statement", [])))
    ]
    return policy


def extract_versions_from_image_name(image_name: str) -> Tuple[str, str]:
    """Returns the python version and ray extracted from an image tag.
    This should be used when creating BYOD images.

    Args:
        image_name: e.g. anyscale/ray-ml:1.11.1-py38-gpu

    Returns:
        The (python version, ray version), e.g. ("py38", "1.11.1")
    """
    # e.g. 1.11.1-py32-gpu
    image_version = image_name.split(":")[-1]
    parts = image_version.split("-")
    # e.g. 1.11.1

    if len(parts) < 2:
        raise ValueError(
            f"Expected the docker image name have an image version tag (something like ray-ml:1.11.1-py38-gpu), got {image_version}."
        )

    ray_version = parts[0]
    # Verify ray_version is valid.
    _ray_version_major_minor(ray_version)

    python_version = parts[1]
    _check_python_version(python_version)
    return (python_version, ray_version)


def _ray_version_major_minor(ray_version: str) -> Tuple[int, int]:
    """Takes in a Ray version such as "1.9.0rc1", "1.10.2", "2.0.0dev0".

    Returns the major minor pair e.g. (1,9) (1,10) (2,0).

    To avoid introducing undesirable dependencies, partly duplicates logic from the Anyscale
    backend.
    """
    invalid_ray_version_msg = (
        f"The Ray version `{ray_version}` has an unexpected format."
    )
    version_components = ray_version.split(".")
    assert len(version_components) >= 2, invalid_ray_version_msg
    major_str, minor_str = version_components[:2]
    assert major_str.isnumeric() and minor_str.isnumeric(), invalid_ray_version_msg
    major_int, minor_int = int(major_str), int(minor_str)
    return (major_int, minor_int)


def _check_python_version(python_version: str) -> None:
    assert (
        python_version in VALID_BYOD_PYTHON_VERSIONS
    ), f"Expected python_version to be one of {VALID_BYOD_PYTHON_VERSIONS}, got {python_version}."


def sleep_till(wake_time: float) -> None:
    """Sleep till the designated time"""
    now = time.time()
    if now >= wake_time:
        return
    time.sleep(wake_time - now)


def poll(
    interval_secs: float = 1,
    timeout_secs: Optional[float] = None,
    max_iter: Optional[int] = None,
) -> Generator[int, None, None]:
    """Poll every interval_secs, until timeout_secs, or max iterations has been reached.
    Yield the iteration number, starting at 1.
    """
    now = time.time()
    end_time = now + timeout_secs if timeout_secs else None
    count = 0
    should_continue_iter = max_iter is None or count < max_iter
    should_continue_time = end_time is None or now < end_time
    while should_continue_iter and should_continue_time:
        count += 1
        wake_time = now + interval_secs
        yield count
        sleep_till(wake_time)
        now = time.time()
        should_continue_iter = max_iter is None or count < max_iter
        should_continue_time = end_time is None or now < end_time


def is_anyscale_workspace() -> bool:
    return "ANYSCALE_EXPERIMENTAL_WORKSPACE_ID" in os.environ


def is_anyscale_cluster() -> bool:
    return "ANYSCALE_SESSION_ID" in os.environ


def credentials_check_sanity(credentials_str: str) -> bool:
    """
    The main goal of this function is to perform a minimal sanity check
    to make sure that the CLI token string (entered by user or read from file)
    is not totally broken.
    refer to https://www.notion.so/anyscale-hq/Authentication-Infrastructure
    """
    # Old token style
    if credentials_str.startswith("sss_"):
        return True
    # Future token style
    return bool(credentials_str.startswith("a") and credentials_str.count("_") > 0)


def get_current_cluster_id() -> Optional[str]:
    """If we are running on an Anyscale Cluster, return the id from the environment."""
    return os.getenv("ANYSCALE_SESSION_ID")


def str_data_size(s: str) -> int:
    """Returns the size of the string when encoded to raw bytes."""
    return len(s.encode("utf-8"))


def get_user_env_aws_account(region: str) -> str:
    """Get the AWS account used in the user environment"""
    return boto3.client("sts", region_name=region).get_caller_identity()["Account"]


def generate_inline_policy_parameter(policy: AnyscaleIAMPolicy) -> str:
    """Generate the inline policy paramter for the cross account role"""
    return f"""  {policy.parameter_key}:
    Description: {policy.parameter_description}
    Type: String"""


def generate_inline_policy_resource(policy: AnyscaleIAMPolicy) -> str:
    """Generate the inline policy resource for the cross account role"""
    return f"""  {policy.resource_logical_id}:
    Type: AWS::IAM::Policy
    Properties:
      PolicyDocument: !Ref {policy.parameter_key}
      PolicyName: {policy.policy_name}
      Roles:
        - !Ref customerRole"""


def get_anyscale_cross_account_iam_policies() -> List[Dict[str, str]]:
    return [
        {
            "ParameterKey": policy.parameter_key,
            "ParameterValue": policy.policy_document,
        }
        for policy in ANYSCALE_IAM_POLICIES
    ]


def prepare_cloudformation_template(
    region: str,
    cfn_stack_name: str,
    cloud_id: str,
    enable_head_node_fault_tolerance: bool,
    boto3_session: Optional[boto3.Session] = None,
    is_anyscale_hosted: bool = False,
    shared_storage: SharedStorageType = SharedStorageType.OBJECT_STORAGE,
) -> str:
    if is_anyscale_hosted:
        with open(f"{anyscale.conf.ROOT_DIR_PATH}/anyscale-cloud-setup-oa.yaml") as f:
            body = f.read()
    else:
        with open(f"{anyscale.conf.ROOT_DIR_PATH}/anyscale-cloud-setup.yaml") as f:
            body = f.read()
        body = body.replace(
            "$ALLOWED_ORIGIN", shared_anyscale_conf.ANYSCALE_CORS_ORIGIN
        )

    zone_ids_to_names = get_availability_zones(region, boto3_session)
    azs = sorted(zone_ids_to_names.values())
    subnet_templates: List[str] = []
    subnets_route_table_association: List[str] = []
    subnets_with_availability_zones: List[str] = []
    efs_mount_targets: List[str] = []

    vpc_cidr = ipaddress.ip_network(VPC_CIDR_RANGE)
    if len(azs) > 4:
        subnet_cidrs = list(vpc_cidr.subnets(prefixlen_diff=3))
    else:
        subnet_cidrs = list(vpc_cidr.subnets(prefixlen_diff=2))

    for i, az in enumerate(azs):
        subnet_templates.append(
            f"""
  Subnet{i}:
    Type: AWS::EC2::Subnet
    Properties:
        VpcId: !Ref VPC
        AvailabilityZone: {az}
        CidrBlock: {subnet_cidrs[i]}
        MapPublicIpOnLaunch: true
        Tags:
        - Key: Name
          Value: {cfn_stack_name}-subnet-{az}
        - Key: anyscale-cloud-id
          Value: {cloud_id}"""
        )

        subnets_route_table_association.append(
            f"""
  Subnet{i}RouteTableAssociation:
    Type: AWS::EC2::SubnetRouteTableAssociation
    Properties:
        RouteTableId: !Ref PublicRouteTable
        SubnetId: !Ref Subnet{i}"""
        )

        subnets_with_availability_zones.append(
            f'{{"subnet_id": !Ref Subnet{i}, "availability_zone": !GetAtt Subnet{i}.AvailabilityZone}}'
        )

        if not is_anyscale_hosted and shared_storage == SharedStorageType.NFS:
            efs_mount_targets.append(
                f"""
  EFSMountTarget{i}:
    Type: AWS::EFS::MountTarget
    Properties:
        FileSystemId: !Ref EFS
        SecurityGroups:
          - !Ref AnyscaleSecurityGroup
        SubnetId: !Ref Subnet{i}"""
            )

    body = body.replace("$SUBNETS_TEMPLATES", "\n".join(subnet_templates))
    body = body.replace(
        "$SUBNETS_ROUTE_TABLE_ASSOCIATION", "\n".join(subnets_route_table_association),
    )
    body = body.replace(
        "$SUBNETS_WITH_AVAILABILITY_ZONES",
        f'[{",".join(subnets_with_availability_zones)}]',
    )
    if not is_anyscale_hosted:
        body = body.replace("$EFSMountTargets", "\n".join(efs_mount_targets))

        # Only include EFS mount target output if mount targets are actually created
        if efs_mount_targets and shared_storage == SharedStorageType.NFS:
            efs_mount_target_output = """
  EFSMountTargetIP:
    Description: Anyscale managed EFS mount target
    Value: !GetAtt
      - EFSMountTarget0
      - IpAddress"""
            body = body.replace("$EFS_MOUNT_TARGET_OUTPUT", efs_mount_target_output)
        else:
            body = body.replace("$EFS_MOUNT_TARGET_OUTPUT", "")
    else:
        # For anyscale hosted, remove the EFS mount target output placeholder
        body = body.replace("$EFS_MOUNT_TARGET_OUTPUT", "")

    iam_policy_parameters = [
        generate_inline_policy_parameter(policy) for policy in ANYSCALE_IAM_POLICIES
    ]
    body = body.replace(
        "$ANYSCALE_CROSS_ACCOUNT_IAM_POLICY_PARAMETERS",
        "\n\n".join(iam_policy_parameters),
    )

    iam_policy_resources = [
        generate_inline_policy_resource(policy) for policy in ANYSCALE_IAM_POLICIES
    ]
    body = body.replace(
        "$ANYSCALE_CROSS_ACCOUNT_IAM_POLICY_RESOURCES",
        "\n\n".join(iam_policy_resources),
    )

    if enable_head_node_fault_tolerance:
        supported_zone_names = get_memorydb_supported_zones(region, zone_ids_to_names)
        allowed_az_indices = sorted(azs.index(az) for az in supported_zone_names)
        body = body.replace("$MEMORY_DB_OUTPUT", MEMORY_DB_OUTPUT,)
        body = body.replace(
            "$MEMORY_DB_RESOURCE",
            MEMORY_DB_RESOURCE.format(
                "\n".join([f"        - !Ref Subnet{i}" for i in allowed_az_indices])
            ),
        )
    else:
        body = body.replace("$MEMORY_DB_OUTPUT", "")
        body = body.replace("$MEMORY_DB_RESOURCE", "")

    return body


def get_latest_ray_version():
    """
    Gets latest Ray version from PYPI. This method should not
    assume Ray is already installed.
    """
    try:
        response = requests.get("https://pypi.org/pypi/ray/json")
        latest_version = response.json()["info"]["version"]
    except Exception as e:  # noqa: BLE001
        log.debug(
            f"Unable to get latest Ray version from https://pypi.org/pypi/ray/json {e!s}"
        )
        latest_version = DEFAULT_RAY_VERSION
    return latest_version


def get_ray_and_py_version_for_default_cluster_env() -> Tuple[str, str]:
    py_version = "".join(str(x) for x in sys.version_info[0:2])
    try:
        import ray  # noqa: PLC0415 - codex_reason("gpt5.2", "optional Ray dependency for local version check")

        ray_version = ray.__version__
        if version.parse(ray_version) < version.parse(MINIMUM_RAY_VERSION):
            raise ValueError(
                f"No default cluster env for Ray version {ray_version}. Please upgrade "
                f'to a version >= {MINIMUM_RAY_VERSION} with `pip install "ray>={MINIMUM_RAY_VERSION}"`.'
            )
        if "dev0" in ray_version:
            raise ValueError(
                f"Your locally installed Ray version is {ray_version}. "
                "There is no default cluster environments for nightly versions of Ray."
            )
    except ImportError:
        # Use latest Ray version if Ray not locally installed, because Anyscale cluster envs for
        # new Ray versions are available before the open source package is released on PYPI.
        ray_version = get_latest_ray_version()
        log.debug(
            f"Ray is not installed locally. Using latest Ray version {ray_version} for "
            "the cluster env."
        )

    return ray_version, py_version


def validate_job_config_dict(
    config_dict: Dict[str, Any], api_client: ProductApi
) -> None:
    """
    Throws an exception if there are invalid values in the config dict.
    """
    compute_config: Optional[Union[ComputeTemplate, DecoratedComputeTemplate]] = None
    if "compute_config" in config_dict and isinstance(
        config_dict["compute_config"], str
    ):
        compute_config = get_cluster_compute_from_name(
            config_dict["compute_config"],
            api_client,
            cloud_name=config_dict.get("cloud"),
        )
    elif "compute_config_id" in config_dict:
        cluster_compute_id = config_dict["compute_config_id"]
        compute_config = api_client.get_compute_template_api_v2_compute_templates_template_id_get(
            cluster_compute_id
        ).result

    if compute_config and compute_config.archived_at:
        raise click.ClickException(
            "This job is using an archived compute config. To submit this job, use another compute config."
        )


# User-facing job states for --v2 mode (maps to multiple backend HaJobStates)
ALLOWED_USER_JOB_STATES = ["STARTING", "RUNNING", "SUCCEEDED", "FAILED"]


def validate_list_jobs_state_filter(_, param, value) -> List[str]:  # noqa: ARG001
    """
    Validate the job state filter for list jobs CLI method.

    Accepts both user-facing states (STARTING, RUNNING, SUCCEEDED, FAILED) for --v2
    and backend HaJobStates (SUCCESS, PENDING, etc.) for legacy compatibility.
    """
    if not value:
        return []
    for each_value in value:
        upper = each_value.upper()
        # Accept both user-facing states (for v2) and backend states (for legacy)
        if (
            upper not in ALLOWED_USER_JOB_STATES
            and upper not in HaJobStates.allowable_values
        ):
            raise click.ClickException(
                f"{each_value} is not valid for {param.opts[0]}. "
                f"For --v2 use: {', '.join(ALLOWED_USER_JOB_STATES)}"
            )
    return [each_value.upper() for each_value in value]


def get_cluster_model_for_current_workspace(
    anyscale_api_client: BaseApi,
) -> Optional[Cluster]:
    """If run within a workspace, returns the `Cluster` info for that workspace.

    Else, returns `None`.
    """
    session_id = os.environ.get("ANYSCALE_SESSION_ID")
    if not session_id or not is_anyscale_workspace():
        return None

    return anyscale_api_client.get_cluster(session_id).result


def populate_unspecified_cluster_configs(
    config: Dict[str, Any], workspace_cluster: Cluster, *, populate_name: bool = False
):
    """Populates unspecified fields in the config from a workspace cluster model.

    This is used to fill in smart defaults when deploying from a workspace.

    Fields that are defaulted:
        - `project_id` if it is not specified.
        - `build_id` if none of {`build_id, `cluster_env`} are specified.
        - `compute_config_id` if none of {`cloud`, `compute_config`, `compute_config_id`} are specified.
        - `name` if `populate_name` is passed and it is not specified.

    The defaulted name will be the name of the wor
    """
    if "project_id" not in config and "project" not in config:
        config["project_id"] = workspace_cluster.project_id

    if "build_id" not in config and "cluster_env" not in config:
        config["build_id"] = workspace_cluster.cluster_environment_build_id

    if (
        "compute_config" not in config
        and "compute_config_id" not in config
        and "cloud" not in config
    ):
        config["compute_config_id"] = workspace_cluster.cluster_compute_id

    if populate_name and "name" not in config:
        name = workspace_cluster.name
        # All workspace cluster names should start with this prefix.
        # Defensively default to the name as-is if they don't.
        if name.startswith("workspace-cluster-"):
            name = name[len("workspace-cluster-") :]

        config["name"] = name

    return config


def populate_unspecified_cluster_configs_from_current_workspace(
    config: Dict[str, Any], anyscale_api_client: BaseApi, *, populate_name: bool = False
) -> Dict[str, Any]:
    workspace_cluster = get_cluster_model_for_current_workspace(anyscale_api_client)
    if workspace_cluster is not None:
        config = populate_unspecified_cluster_configs(
            config, workspace_cluster, populate_name=populate_name
        )

    return config


def filter_actions_from_policy_document(
    policy_document: Dict[Any, Any],
    action_filter: Optional[Callable[[Dict], bool]] = None,
) -> set:
    if action_filter is None:
        action_filter = lambda statement: statement["Effect"] == "Allow"  # noqa: E731
    return {
        action
        for statement in _coerce_to_list(policy_document.get("Statement", {}))
        for action in _coerce_to_list(statement.get("Action"))
        if action_filter(statement)
    }


def filter_actions_associated_with_role(
    boto3_session: boto3.Session,
    role: Boto3Resource,
    action_filter: Optional[Callable[[Dict], bool]] = None,
) -> Set[str]:
    iam = boto3_session.resource("iam")
    attached_policy_documents = [
        iam.PolicyVersion(policy.arn, policy.default_version_id).document
        for policy in role.attached_policies.all()  # type: ignore
    ]

    role_policy_documents = [policy.policy_document for policy in role.policies.all()]  # type: ignore

    list_of_allow_actions_sets = [
        filter_actions_from_policy_document(policy_document, action_filter)
        for policy_document in role_policy_documents + attached_policy_documents
    ]
    return (
        set.union(*list_of_allow_actions_sets) if list_of_allow_actions_sets else set()
    )


def contains_control_plane_role(
    assume_role_policy_document: Dict[str, Any], anyscale_aws_account: str
) -> bool:
    def action_filter(statement: dict):
        if not statement:
            return False
        # Ensure it is `Allow` & `sts:AssumeRole`
        if (
            statement["Effect"] != "Allow"
            or statement.get("Action") != "sts:AssumeRole"
        ):
            return False

        expected_accounts = {
            f"arn:aws:iam::{anyscale_aws_account}:root",
            anyscale_aws_account,
        }
        return any(
            aws_principal in expected_accounts
            for aws_principal in _coerce_to_list(statement["Principal"].get("AWS"))
        )

    return (
        len(
            filter_actions_from_policy_document(
                assume_role_policy_document, action_filter
            )
        )
        > 0
    )


def verify_data_plane_role_assume_role_policy(
    assume_role_policy_document: Dict[str, Any]
) -> bool:
    def action_filter(statement: dict):
        # Ensure it is `Allow` & `sts:AssumeRole`
        if not statement:
            return False
        if (
            statement["Effect"] != "Allow"
            or statement.get("Action") != "sts:AssumeRole"
        ):
            return False

        return any(
            service == "ec2.amazonaws.com"
            for service in _coerce_to_list(statement["Principal"].get("Service"))
        )

    return (
        len(
            filter_actions_from_policy_document(
                assume_role_policy_document, action_filter
            )
        )
        > 0
    )


T = TypeVar("T")


def _coerce_to_list(maybe_list: Union[T, List[T]]) -> List[T]:
    return maybe_list if isinstance(maybe_list, list) else [maybe_list]


def allow_optional_file_storage(api_client: Optional[ProductApi] = None) -> bool:
    if api_client is None:
        api_client = get_auth_api_client().api_client
    return api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get(
        "allow-optional-file-storage"
    ).result.is_on



"""CORS update utilities for cloud storage buckets.

This module provides utilities to check and update CORS configurations
on cloud storage buckets (S3, GCS, Azure Blob) to support Anyscale UI
features like the file viewer.
"""

from typing import List, Optional, Tuple, TYPE_CHECKING


if TYPE_CHECKING:
    import boto3

from anyscale.cloud_resource import check_aws_cors_rules
from anyscale.shared_anyscale_utils.conf import (
    ANYSCALE_CORS_EXPOSE_HEADERS,
    ANYSCALE_CORS_ORIGIN,
)


def check_aws_cors_needs_update(
    bucket_name: str, region: str, boto3_session: Optional["boto3.Session"] = None
) -> Tuple[bool, str]:
    """
    Check if AWS S3 bucket CORS needs update for file viewer support.

    Args:
        bucket_name: The S3 bucket name (without s3:// prefix)
        region: AWS region
        boto3_session: Optional boto3 session to use

    Returns:
        Tuple of (needs_update, reason)
    """
    try:
        import boto3 as boto3_module  # noqa: PLC0415
        from botocore.exceptions import ClientError  # noqa: PLC0415
    except ImportError as e:
        raise ImportError(
            "AWS SDK is not installed. Please install it with: pip install boto3"
        ) from e

    if boto3_session is None:
        boto3_session = boto3_module.Session()

    s3 = boto3_session.resource("s3", region_name=region)
    bucket = s3.Bucket(bucket_name)

    try:
        cors_rules = bucket.Cors().cors_rules
    except ClientError as e:
        if e.response["Error"]["Code"] == "NoSuchCORSConfiguration":
            return True, "No CORS configuration exists"
        raise

    is_correct, reason = check_aws_cors_rules(cors_rules)
    return not is_correct, reason


def update_aws_cors(
    bucket_name: str, region: str, boto3_session: Optional["boto3.Session"] = None
) -> None:
    """
    Update AWS S3 bucket CORS configuration.

    Args:
        bucket_name: The S3 bucket name (without s3:// prefix)
        region: AWS region
        boto3_session: Optional boto3 session to use
    """
    try:
        import boto3 as boto3_module  # noqa: PLC0415
    except ImportError as e:
        raise ImportError(
            "AWS SDK is not installed. Please install it with: pip install boto3"
        ) from e

    if boto3_session is None:
        boto3_session = boto3_module.Session()

    s3_client = boto3_session.client("s3", region_name=region)
    s3_client.put_bucket_cors(
        Bucket=bucket_name,
        CORSConfiguration={
            "CORSRules": [
                {
                    "AllowedHeaders": ["*"],
                    "AllowedMethods": ["GET", "PUT", "POST", "HEAD", "DELETE"],
                    "AllowedOrigins": [ANYSCALE_CORS_ORIGIN],
                    "ExposeHeaders": ANYSCALE_CORS_EXPOSE_HEADERS,
                    "MaxAgeSeconds": 3600,
                }
            ]
        },
    )


def check_gcp_cors_needs_update(
    bucket_name: str, project_id: Optional[str]
) -> Tuple[bool, str]:
    """
    Check if GCP GCS bucket CORS needs update.

    Args:
        bucket_name: The GCS bucket name (without gs:// prefix)
        project_id: GCP project ID

    Returns:
        Tuple of (needs_update, reason)
    """
    try:
        from google.cloud import storage  # noqa: PLC0415

        from anyscale.gcp_verification import check_gcp_cors_rules  # noqa: PLC0415
    except ImportError as e:
        raise ImportError(
            "Google Cloud SDK is not installed. Please install it with: "
            "pip install google-cloud-storage"
        ) from e

    client = storage.Client(project=project_id)
    bucket = client.get_bucket(bucket_name)

    is_correct, reason = check_gcp_cors_rules(bucket.cors or [])
    return not is_correct, reason


def update_gcp_cors(bucket_name: str, project_id: Optional[str]) -> None:
    """
    Update GCP GCS bucket CORS configuration.

    Args:
        bucket_name: The GCS bucket name (without gs:// prefix)
        project_id: GCP project ID
    """
    try:
        from google.cloud import storage  # noqa: PLC0415
    except ImportError as e:
        raise ImportError(
            "Google Cloud SDK is not installed. Please install it with: "
            "pip install google-cloud-storage"
        ) from e

    client = storage.Client(project=project_id)
    bucket = client.get_bucket(bucket_name)
    bucket.cors = [
        {
            "origin": [ANYSCALE_CORS_ORIGIN],
            "responseHeader": ["*"],
            "method": ["GET", "PUT", "POST", "HEAD", "DELETE"],
            "maxAgeSeconds": 3600,
        }
    ]
    bucket.patch()


def check_azure_cors_rules(cors_rules: List) -> Tuple[bool, str]:
    """
    Check if Azure Blob storage CORS rules are correctly configured for Anyscale.

    This is a shared helper used by both CORS update utilities and cloud verification.

    Args:
        cors_rules: List of CorsRule objects from Azure Blob storage

    Returns:
        Tuple of (is_correct, reason)
    """
    if not cors_rules:
        return False, "No CORS configuration exists"

    for rule in cors_rules:
        if ANYSCALE_CORS_ORIGIN in rule.allowed_origins and "*" in rule.exposed_headers:
            return True, "CORS already configured correctly"

    return False, "CORS missing required exposedHeaders for file viewer"


def check_azure_cors_needs_update(storage_account_name: str) -> Tuple[bool, str]:
    """
    Check if Azure Blob storage CORS needs update.

    Args:
        storage_account_name: Azure storage account name

    Returns:
        Tuple of (needs_update, reason)
    """
    try:
        from azure.identity import DefaultAzureCredential  # noqa: PLC0415
        from azure.storage.blob import BlobServiceClient  # noqa: PLC0415
    except ImportError as e:
        raise ImportError(
            "Azure SDK is not installed. Please install it with: "
            "pip install azure-identity azure-storage-blob"
        ) from e

    credential = DefaultAzureCredential()
    account_url = f"https://{storage_account_name}.blob.core.windows.net"
    blob_service = BlobServiceClient(account_url=account_url, credential=credential)

    props = blob_service.get_service_properties()
    cors_rules = props.get("cors", [])

    is_correct, reason = check_azure_cors_rules(cors_rules)
    return not is_correct, reason


def update_azure_cors(storage_account_name: str) -> None:
    """
    Update Azure Blob storage CORS configuration.

    Args:
        storage_account_name: Azure storage account name
    """
    try:
        from azure.identity import DefaultAzureCredential  # noqa: PLC0415
        from azure.storage.blob import BlobServiceClient, CorsRule  # noqa: PLC0415
    except ImportError as e:
        raise ImportError(
            "Azure SDK is not installed. Please install it with: "
            "pip install azure-identity azure-storage-blob"
        ) from e

    credential = DefaultAzureCredential()
    account_url = f"https://{storage_account_name}.blob.core.windows.net"
    blob_service = BlobServiceClient(account_url=account_url, credential=credential)

    cors_rule = CorsRule(
        allowed_origins=[ANYSCALE_CORS_ORIGIN],
        allowed_methods=["GET", "HEAD", "PUT"],
        allowed_headers=["*"],
        exposed_headers=["*"],
        max_age_in_seconds=3600,
    )
    blob_service.set_service_properties(cors=[cors_rule])



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class DataCatalogTable(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'name': 'str',
        'description': 'str',
        'created_at': 'str',
        'updated_at': 'str',
        'schema_id': 'str',
        'catalog_id': 'str',
        'connection_id': 'str',
        'object_type': 'DataCatalogObjectType',
        'table_type': 'str',
        'data_format': 'str',
        'metadata': 'BackendServerApiProductModelsDataCatalogsTableMetadata'
    }

    attribute_map = {
        'id': 'id',
        'name': 'name',
        'description': 'description',
        'created_at': 'created_at',
        'updated_at': 'updated_at',
        'schema_id': 'schema_id',
        'catalog_id': 'catalog_id',
        'connection_id': 'connection_id',
        'object_type': 'object_type',
        'table_type': 'table_type',
        'data_format': 'data_format',
        'metadata': 'metadata'
    }

    def __init__(self, id=None, name=None, description=None, created_at=None, updated_at=None, schema_id=None, catalog_id=None, connection_id=None, object_type=None, table_type=None, data_format=None, metadata=None, local_vars_configuration=None):  # noqa: E501
        """DataCatalogTable - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._name = None
        self._description = None
        self._created_at = None
        self._updated_at = None
        self._schema_id = None
        self._catalog_id = None
        self._connection_id = None
        self._object_type = None
        self._table_type = None
        self._data_format = None
        self._metadata = None
        self.discriminator = None

        self.id = id
        self.name = name
        if description is not None:
            self.description = description
        self.created_at = created_at
        self.updated_at = updated_at
        self.schema_id = schema_id
        self.catalog_id = catalog_id
        self.connection_id = connection_id
        if object_type is not None:
            self.object_type = object_type
        if table_type is not None:
            self.table_type = table_type
        if data_format is not None:
            self.data_format = data_format
        if metadata is not None:
            self.metadata = metadata

    @property
    def id(self):
        """Gets the id of this DataCatalogTable.  # noqa: E501

        Unique identifier  # noqa: E501

        :return: The id of this DataCatalogTable.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this DataCatalogTable.

        Unique identifier  # noqa: E501

        :param id: The id of this DataCatalogTable.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def name(self):
        """Gets the name of this DataCatalogTable.  # noqa: E501

        Display name  # noqa: E501

        :return: The name of this DataCatalogTable.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this DataCatalogTable.

        Display name  # noqa: E501

        :param name: The name of this DataCatalogTable.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def description(self):
        """Gets the description of this DataCatalogTable.  # noqa: E501

        Description  # noqa: E501

        :return: The description of this DataCatalogTable.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this DataCatalogTable.

        Description  # noqa: E501

        :param description: The description of this DataCatalogTable.  # noqa: E501
        :type: str
        """

        self._description = description

    @property
    def created_at(self):
        """Gets the created_at of this DataCatalogTable.  # noqa: E501

        Creation timestamp in ISO 8601 format  # noqa: E501

        :return: The created_at of this DataCatalogTable.  # noqa: E501
        :rtype: str
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this DataCatalogTable.

        Creation timestamp in ISO 8601 format  # noqa: E501

        :param created_at: The created_at of this DataCatalogTable.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def updated_at(self):
        """Gets the updated_at of this DataCatalogTable.  # noqa: E501

        Last update timestamp in ISO 8601 format  # noqa: E501

        :return: The updated_at of this DataCatalogTable.  # noqa: E501
        :rtype: str
        """
        return self._updated_at

    @updated_at.setter
    def updated_at(self, updated_at):
        """Sets the updated_at of this DataCatalogTable.

        Last update timestamp in ISO 8601 format  # noqa: E501

        :param updated_at: The updated_at of this DataCatalogTable.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and updated_at is None:  # noqa: E501
            raise ValueError("Invalid value for `updated_at`, must not be `None`")  # noqa: E501

        self._updated_at = updated_at

    @property
    def schema_id(self):
        """Gets the schema_id of this DataCatalogTable.  # noqa: E501

        ID of the parent schema  # noqa: E501

        :return: The schema_id of this DataCatalogTable.  # noqa: E501
        :rtype: str
        """
        return self._schema_id

    @schema_id.setter
    def schema_id(self, schema_id):
        """Sets the schema_id of this DataCatalogTable.

        ID of the parent schema  # noqa: E501

        :param schema_id: The schema_id of this DataCatalogTable.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and schema_id is None:  # noqa: E501
            raise ValueError("Invalid value for `schema_id`, must not be `None`")  # noqa: E501

        self._schema_id = schema_id

    @property
    def catalog_id(self):
        """Gets the catalog_id of this DataCatalogTable.  # noqa: E501

        ID of the parent catalog  # noqa: E501

        :return: The catalog_id of this DataCatalogTable.  # noqa: E501
        :rtype: str
        """
        return self._catalog_id

    @catalog_id.setter
    def catalog_id(self, catalog_id):
        """Sets the catalog_id of this DataCatalogTable.

        ID of the parent catalog  # noqa: E501

        :param catalog_id: The catalog_id of this DataCatalogTable.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and catalog_id is None:  # noqa: E501
            raise ValueError("Invalid value for `catalog_id`, must not be `None`")  # noqa: E501

        self._catalog_id = catalog_id

    @property
    def connection_id(self):
        """Gets the connection_id of this DataCatalogTable.  # noqa: E501

        ID of the parent data catalog connection  # noqa: E501

        :return: The connection_id of this DataCatalogTable.  # noqa: E501
        :rtype: str
        """
        return self._connection_id

    @connection_id.setter
    def connection_id(self, connection_id):
        """Sets the connection_id of this DataCatalogTable.

        ID of the parent data catalog connection  # noqa: E501

        :param connection_id: The connection_id of this DataCatalogTable.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and connection_id is None:  # noqa: E501
            raise ValueError("Invalid value for `connection_id`, must not be `None`")  # noqa: E501

        self._connection_id = connection_id

    @property
    def object_type(self):
        """Gets the object_type of this DataCatalogTable.  # noqa: E501

        Type of catalog object (always 'table' for this model)  # noqa: E501

        :return: The object_type of this DataCatalogTable.  # noqa: E501
        :rtype: DataCatalogObjectType
        """
        return self._object_type

    @object_type.setter
    def object_type(self, object_type):
        """Sets the object_type of this DataCatalogTable.

        Type of catalog object (always 'table' for this model)  # noqa: E501

        :param object_type: The object_type of this DataCatalogTable.  # noqa: E501
        :type: DataCatalogObjectType
        """

        self._object_type = object_type

    @property
    def table_type(self):
        """Gets the table_type of this DataCatalogTable.  # noqa: E501

        Type of table (e.g., 'MANAGED', 'EXTERNAL', 'VIEW')  # noqa: E501

        :return: The table_type of this DataCatalogTable.  # noqa: E501
        :rtype: str
        """
        return self._table_type

    @table_type.setter
    def table_type(self, table_type):
        """Sets the table_type of this DataCatalogTable.

        Type of table (e.g., 'MANAGED', 'EXTERNAL', 'VIEW')  # noqa: E501

        :param table_type: The table_type of this DataCatalogTable.  # noqa: E501
        :type: str
        """

        self._table_type = table_type

    @property
    def data_format(self):
        """Gets the data_format of this DataCatalogTable.  # noqa: E501

        Data format (e.g., 'DELTA', 'PARQUET')  # noqa: E501

        :return: The data_format of this DataCatalogTable.  # noqa: E501
        :rtype: str
        """
        return self._data_format

    @data_format.setter
    def data_format(self, data_format):
        """Sets the data_format of this DataCatalogTable.

        Data format (e.g., 'DELTA', 'PARQUET')  # noqa: E501

        :param data_format: The data_format of this DataCatalogTable.  # noqa: E501
        :type: str
        """

        self._data_format = data_format

    @property
    def metadata(self):
        """Gets the metadata of this DataCatalogTable.  # noqa: E501

        Table-specific metadata including owner, full_name, and column_count  # noqa: E501

        :return: The metadata of this DataCatalogTable.  # noqa: E501
        :rtype: BackendServerApiProductModelsDataCatalogsTableMetadata
        """
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        """Sets the metadata of this DataCatalogTable.

        Table-specific metadata including owner, full_name, and column_count  # noqa: E501

        :param metadata: The metadata of this DataCatalogTable.  # noqa: E501
        :type: BackendServerApiProductModelsDataCatalogsTableMetadata
        """

        self._metadata = metadata

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DataCatalogTable):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DataCatalogTable):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class DecoratedJobConnectionInfo(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'workload_id': 'str',
        'workload_type': 'str',
        'connection': 'ConnectionInfo',
        'created_at': 'datetime',
        'created_by': 'MiniUser',
        'deleted_at': 'datetime',
        'deleted_by': 'MiniUser'
    }

    attribute_map = {
        'id': 'id',
        'workload_id': 'workload_id',
        'workload_type': 'workload_type',
        'connection': 'connection',
        'created_at': 'created_at',
        'created_by': 'created_by',
        'deleted_at': 'deleted_at',
        'deleted_by': 'deleted_by'
    }

    def __init__(self, id=None, workload_id=None, workload_type=None, connection=None, created_at=None, created_by=None, deleted_at=None, deleted_by=None, local_vars_configuration=None):  # noqa: E501
        """DecoratedJobConnectionInfo - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._workload_id = None
        self._workload_type = None
        self._connection = None
        self._created_at = None
        self._created_by = None
        self._deleted_at = None
        self._deleted_by = None
        self.discriminator = None

        self.id = id
        self.workload_id = workload_id
        self.workload_type = workload_type
        self.connection = connection
        self.created_at = created_at
        self.created_by = created_by
        if deleted_at is not None:
            self.deleted_at = deleted_at
        if deleted_by is not None:
            self.deleted_by = deleted_by

    @property
    def id(self):
        """Gets the id of this DecoratedJobConnectionInfo.  # noqa: E501

        Workload connection ID  # noqa: E501

        :return: The id of this DecoratedJobConnectionInfo.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this DecoratedJobConnectionInfo.

        Workload connection ID  # noqa: E501

        :param id: The id of this DecoratedJobConnectionInfo.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def workload_id(self):
        """Gets the workload_id of this DecoratedJobConnectionInfo.  # noqa: E501

        Workload (job, workspace, etc) ID  # noqa: E501

        :return: The workload_id of this DecoratedJobConnectionInfo.  # noqa: E501
        :rtype: str
        """
        return self._workload_id

    @workload_id.setter
    def workload_id(self, workload_id):
        """Sets the workload_id of this DecoratedJobConnectionInfo.

        Workload (job, workspace, etc) ID  # noqa: E501

        :param workload_id: The workload_id of this DecoratedJobConnectionInfo.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and workload_id is None:  # noqa: E501
            raise ValueError("Invalid value for `workload_id`, must not be `None`")  # noqa: E501

        self._workload_id = workload_id

    @property
    def workload_type(self):
        """Gets the workload_type of this DecoratedJobConnectionInfo.  # noqa: E501

        Type of workload (job, workspace, etc.)  # noqa: E501

        :return: The workload_type of this DecoratedJobConnectionInfo.  # noqa: E501
        :rtype: str
        """
        return self._workload_type

    @workload_type.setter
    def workload_type(self, workload_type):
        """Sets the workload_type of this DecoratedJobConnectionInfo.

        Type of workload (job, workspace, etc.)  # noqa: E501

        :param workload_type: The workload_type of this DecoratedJobConnectionInfo.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and workload_type is None:  # noqa: E501
            raise ValueError("Invalid value for `workload_type`, must not be `None`")  # noqa: E501

        self._workload_type = workload_type

    @property
    def connection(self):
        """Gets the connection of this DecoratedJobConnectionInfo.  # noqa: E501

        Connection details  # noqa: E501

        :return: The connection of this DecoratedJobConnectionInfo.  # noqa: E501
        :rtype: ConnectionInfo
        """
        return self._connection

    @connection.setter
    def connection(self, connection):
        """Sets the connection of this DecoratedJobConnectionInfo.

        Connection details  # noqa: E501

        :param connection: The connection of this DecoratedJobConnectionInfo.  # noqa: E501
        :type: ConnectionInfo
        """
        if self.local_vars_configuration.client_side_validation and connection is None:  # noqa: E501
            raise ValueError("Invalid value for `connection`, must not be `None`")  # noqa: E501

        self._connection = connection

    @property
    def created_at(self):
        """Gets the created_at of this DecoratedJobConnectionInfo.  # noqa: E501

        When the connection was associated  # noqa: E501

        :return: The created_at of this DecoratedJobConnectionInfo.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this DecoratedJobConnectionInfo.

        When the connection was associated  # noqa: E501

        :param created_at: The created_at of this DecoratedJobConnectionInfo.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def created_by(self):
        """Gets the created_by of this DecoratedJobConnectionInfo.  # noqa: E501

        User who associated the connection  # noqa: E501

        :return: The created_by of this DecoratedJobConnectionInfo.  # noqa: E501
        :rtype: MiniUser
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this DecoratedJobConnectionInfo.

        User who associated the connection  # noqa: E501

        :param created_by: The created_by of this DecoratedJobConnectionInfo.  # noqa: E501
        :type: MiniUser
        """
        if self.local_vars_configuration.client_side_validation and created_by is None:  # noqa: E501
            raise ValueError("Invalid value for `created_by`, must not be `None`")  # noqa: E501

        self._created_by = created_by

    @property
    def deleted_at(self):
        """Gets the deleted_at of this DecoratedJobConnectionInfo.  # noqa: E501

        When the connection was disassociated  # noqa: E501

        :return: The deleted_at of this DecoratedJobConnectionInfo.  # noqa: E501
        :rtype: datetime
        """
        return self._deleted_at

    @deleted_at.setter
    def deleted_at(self, deleted_at):
        """Sets the deleted_at of this DecoratedJobConnectionInfo.

        When the connection was disassociated  # noqa: E501

        :param deleted_at: The deleted_at of this DecoratedJobConnectionInfo.  # noqa: E501
        :type: datetime
        """

        self._deleted_at = deleted_at

    @property
    def deleted_by(self):
        """Gets the deleted_by of this DecoratedJobConnectionInfo.  # noqa: E501

        User who disassociated the connection  # noqa: E501

        :return: The deleted_by of this DecoratedJobConnectionInfo.  # noqa: E501
        :rtype: MiniUser
        """
        return self._deleted_by

    @deleted_by.setter
    def deleted_by(self, deleted_by):
        """Sets the deleted_by of this DecoratedJobConnectionInfo.

        User who disassociated the connection  # noqa: E501

        :param deleted_by: The deleted_by of this DecoratedJobConnectionInfo.  # noqa: E501
        :type: MiniUser
        """

        self._deleted_by = deleted_by

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DecoratedJobConnectionInfo):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DecoratedJobConnectionInfo):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class DescribeMachinePoolMachinesFilters(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'allocation_states': 'list[MachineAllocationState]',
        'partitions': 'list[str]',
        'machine_type_partition_filters': 'list[MachineTypePartitionFilter]'
    }

    attribute_map = {
        'allocation_states': 'allocation_states',
        'partitions': 'partitions',
        'machine_type_partition_filters': 'machine_type_partition_filters'
    }

    def __init__(self, allocation_states=None, partitions=None, machine_type_partition_filters=None, local_vars_configuration=None):  # noqa: E501
        """DescribeMachinePoolMachinesFilters - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._allocation_states = None
        self._partitions = None
        self._machine_type_partition_filters = None
        self.discriminator = None

        if allocation_states is not None:
            self.allocation_states = allocation_states
        if partitions is not None:
            self.partitions = partitions
        if machine_type_partition_filters is not None:
            self.machine_type_partition_filters = machine_type_partition_filters

    @property
    def allocation_states(self):
        """Gets the allocation_states of this DescribeMachinePoolMachinesFilters.  # noqa: E501

        The allocation states to filter by.  # noqa: E501

        :return: The allocation_states of this DescribeMachinePoolMachinesFilters.  # noqa: E501
        :rtype: list[MachineAllocationState]
        """
        return self._allocation_states

    @allocation_states.setter
    def allocation_states(self, allocation_states):
        """Sets the allocation_states of this DescribeMachinePoolMachinesFilters.

        The allocation states to filter by.  # noqa: E501

        :param allocation_states: The allocation_states of this DescribeMachinePoolMachinesFilters.  # noqa: E501
        :type: list[MachineAllocationState]
        """

        self._allocation_states = allocation_states

    @property
    def partitions(self):
        """Gets the partitions of this DescribeMachinePoolMachinesFilters.  # noqa: E501

        The partitions to filter by (for backward compatibility).  # noqa: E501

        :return: The partitions of this DescribeMachinePoolMachinesFilters.  # noqa: E501
        :rtype: list[str]
        """
        return self._partitions

    @partitions.setter
    def partitions(self, partitions):
        """Sets the partitions of this DescribeMachinePoolMachinesFilters.

        The partitions to filter by (for backward compatibility).  # noqa: E501

        :param partitions: The partitions of this DescribeMachinePoolMachinesFilters.  # noqa: E501
        :type: list[str]
        """

        self._partitions = partitions

    @property
    def machine_type_partition_filters(self):
        """Gets the machine_type_partition_filters of this DescribeMachinePoolMachinesFilters.  # noqa: E501

        The machine_type, partition pair to filter by. If not provided, no filtering is performed.  # noqa: E501

        :return: The machine_type_partition_filters of this DescribeMachinePoolMachinesFilters.  # noqa: E501
        :rtype: list[MachineTypePartitionFilter]
        """
        return self._machine_type_partition_filters

    @machine_type_partition_filters.setter
    def machine_type_partition_filters(self, machine_type_partition_filters):
        """Sets the machine_type_partition_filters of this DescribeMachinePoolMachinesFilters.

        The machine_type, partition pair to filter by. If not provided, no filtering is performed.  # noqa: E501

        :param machine_type_partition_filters: The machine_type_partition_filters of this DescribeMachinePoolMachinesFilters.  # noqa: E501
        :type: list[MachineTypePartitionFilter]
        """

        self._machine_type_partition_filters = machine_type_partition_filters

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DescribeMachinePoolMachinesFilters):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DescribeMachinePoolMachinesFilters):
            return True

        return self.to_dict() != other.to_dict()



from io import StringIO
import json
from typing import IO, Optional, Tuple

import click
import tabulate
import yaml

import anyscale
from anyscale.cli_logger import BlockLogger
from anyscale.cloud_utils import get_cloud_id_and_name
from anyscale.commands import command_examples
from anyscale.commands.util import AnyscaleCommand
from anyscale.compute_config.models import (
    compute_config_type_from_yaml,
    ComputeConfigVersion,
)
from anyscale.controllers.base_controller import BaseController
from anyscale.controllers.compute_config_controller import ComputeConfigController
from anyscale.util import get_endpoint, validate_non_negative_arg


logger = BlockLogger()


def _validate_name_and_id_args(
    *, positional_name: Optional[str], flag_name: Optional[str], id_flag: Optional[str]
) -> Tuple[Optional[str], Optional[str]]:
    """Handles validation and deduplication for the name & ID options.

    The positional name is deprecated and will be removed in the near future.

    Returns (name, ID) -- exactly one of them will be populated.
    """
    if flag_name and positional_name:
        raise click.ClickException(
            "Both -n/--name and [COMPUTE_CONFIG_NAME] were provided. "
            "Use -n/--name only (the positional argument is deprecated)."
        )
    elif positional_name:
        logger.warning(
            "The [COMPUTE_CONFIG_NAME] positional argument is deprecated. "
            "Use the -n/--name flag instead."
        )
        name = positional_name
    elif flag_name:
        name = flag_name
    else:
        name = None

    if name and id_flag:
        raise click.ClickException("Only one of name or ID can be provided.")

    if not name and not id_flag:
        raise click.ClickException(
            "Either -n/--name or --id/--compute-config-id must be provided."
        )

    return name, id_flag


@click.group(
    "compute-config", help="Manage compute configurations.",
)
def compute_config_cli() -> None:
    pass


@compute_config_cli.command(
    name="create",
    help=(
        "Create a new version of a compute config from a YAML file.\n\n"
        "(1) To use the **new schema** defined at "
        "https://docs.anyscale.com/reference/compute-config-api#computeconfig, "
        "use the -f/--config-file flag:\n\n"
        "`anyscale compute-config create -f new_schema_config.yaml`\n\n"
        "(2) To use the **old schema** defined at "
        "https://docs.anyscale.com/reference/compute-config-api/#createclustercomputeconfig-legacy, "
        "use the positional argument:\n\n"
        "`anyscale compute-config create old_schema_config.yaml`\n\n"
    ),
    cls=AnyscaleCommand,
    example=command_examples.COMPUTE_CONFIG_CREATE_EXAMPLE,
)
@click.argument("compute-config-file", type=click.File("rb"), required=False)
@click.option(
    "-n",
    "--name",
    help="Name for the created compute config. This should *not* include a version tag. If a name is not provided, an anonymous compute config is generated. Anonymous compute configs are not accessible in the UI and can only be referenced by their ID.",
    required=False,
    default=None,
    type=str,
)
@click.option(
    "-f",
    "--config-file",
    required=False,
    default=None,
    type=str,
    help="Path to a YAML config file defining the compute config. Schema: https://docs.anyscale.com/reference/compute-config-api#computeconfig.",
)
def create_compute_config(
    compute_config_file: Optional[IO[bytes]],
    config_file: Optional[str],
    name: Optional[str],
):
    if compute_config_file and config_file:
        raise click.ClickException(
            "Only one of the --config-file flag or [COMPUTE_CONFIG_FILE] argument can be provided."
        )

    if compute_config_file is not None:
        ComputeConfigController().create(compute_config_file, name)
    elif config_file is not None:
        config = compute_config_type_from_yaml(config_file)
        anyscale.compute_config.create(config, name=name)
    else:
        raise click.ClickException(
            "Either the --config-file flag or [COMPUTE_CONFIG_FILE] argument must be provided."
        )


@compute_config_cli.command(
    name="archive",
    help=("Archive all versions of a specified compute config.\n\n"),
    cls=AnyscaleCommand,
    example=command_examples.COMPUTE_CONFIG_ARCHIVE_EXAMPLE,
)
@click.argument("compute-config-name", type=str, required=False)
@click.option(
    "-n",
    "--name",
    help="Name of the compute config to archive.",
    required=False,
    type=str,
)
@click.option(
    "--compute-config-id",
    "--id",
    help="ID of the compute config to archive. Alternative to name.",
    required=False,
    type=str,
)
def archive_compute_config(
    compute_config_name: Optional[str],
    name: Optional[str],
    compute_config_id: Optional[str],
) -> None:
    name, cc_id = _validate_name_and_id_args(
        positional_name=compute_config_name, flag_name=name, id_flag=compute_config_id
    )
    anyscale.compute_config.archive(name=name, _id=cc_id)


@compute_config_cli.command(
    name="list",
    help=(
        "List compute configurations with filtering, sorting, and pagination.\n\n"
        "By default, only compute configs created by the current user are returned. "
        "Use --include-shared to include configs shared with you.\n\n"
        "Use --max-items to control page size and --next-token for pagination."
    ),
    cls=AnyscaleCommand,
    example=command_examples.COMPUTE_CONFIG_LIST_EXAMPLE,
)
@click.option(
    "-n",
    "--name",
    required=False,
    default=None,
    help="List information about the compute config with this name.",
)
@click.option(
    "--compute-config-id",
    "--id",
    required=False,
    default=None,
    help="List information about the compute config with this id.",
)
@click.option(
    "--include-shared",
    is_flag=True,
    default=False,
    help="Include all compute configs you have access to.",
)
@click.option(
    "--max-items",
    required=False,
    default=20,
    type=int,
    help="Maximum number of items to return per page (default: 20).",
    callback=validate_non_negative_arg,
)
@click.option(
    "--next-token",
    required=False,
    default=None,
    type=str,
    help="Token for pagination to fetch the next page of results.",
)
@click.option(
    "--cloud-id", required=False, default=None, type=str, help="Filter by cloud ID.",
)
@click.option(
    "--cloud-name",
    required=False,
    default=None,
    type=str,
    help="Filter by cloud name.",
)
@click.option(
    "--sort-by",
    required=False,
    default="last_modified_at",
    type=click.Choice(["name", "created_at", "last_modified_at"], case_sensitive=False),
    help="Field to sort by. Default: last_modified_at",
)
@click.option(
    "--sort-order",
    required=False,
    default="asc",
    type=click.Choice(["asc", "desc"], case_sensitive=False),
    help="Sort order. Default: asc (ascending)",
)
@click.option(
    "--json",
    "output_json",
    is_flag=True,
    default=False,
    help="Output results in JSON format.",
)
def list_compute_configs(  # noqa: A001, PLR0913
    name: Optional[str],
    compute_config_id: Optional[str],
    include_shared: bool,
    max_items: int,
    next_token: Optional[str],
    cloud_id: Optional[str],
    cloud_name: Optional[str],
    sort_by: str,
    sort_order: str,
    output_json: bool,
):
    """List compute configurations with filtering, sorting, and pagination options."""
    # Validate mutual exclusion: cloud_id and cloud_name cannot be used together
    if cloud_id and cloud_name:
        raise click.ClickException(
            "Error: --cloud-id and --cloud-name are mutually exclusive. "
            "Please provide only one."
        )

    # Use the SDK for listing compute configs
    result = anyscale.compute_config.list(
        name=name,
        _id=compute_config_id,
        cloud_id=cloud_id,
        cloud_name=cloud_name,
        sort_by=sort_by,
        sort_order=sort_order,
        max_items=max_items,
        next_token=next_token,
        include_shared=include_shared,
    )

    # Output in JSON format if requested
    if output_json:
        output_data = {
            "results": [
                {
                    "id": cc.id,
                    "name": cc.name,
                    "cloud_id": cc.config.cloud_id if cc.config else None,
                    "version": cc.version,
                    "created_at": cc.created_at.isoformat() if cc.created_at else None,
                    "last_modified_at": cc.last_modified_at.isoformat()
                    if cc.last_modified_at
                    else None,
                    "url": get_endpoint(f"configurations/cluster-computes/{cc.id}"),
                }
                for cc in result.results
            ],
            "metadata": {"count": result.count, "next_token": result.next_token,},
        }
        print(json.dumps(output_data, indent=2))
        return

    # Build table for display
    api_client = BaseController().anyscale_api_client

    cluster_compute_table = [
        [
            cluster_compute.id,
            cluster_compute.name,
            api_client.get_cloud(cluster_compute.config.cloud_id).result.name
            if cluster_compute.config.cloud_id
            else None,
            cluster_compute.last_modified_at.strftime("%m/%d/%Y, %H:%M:%S"),
            get_endpoint(f"configurations/cluster-computes/{cluster_compute.id}"),
        ]
        for cluster_compute in result.results
    ]

    table = tabulate.tabulate(
        cluster_compute_table,
        headers=["ID", "NAME", "CLOUD", "LAST MODIFIED AT", "URL"],
        tablefmt="plain",
    )
    print(f"Compute configs:\n{table}")

    # Print pagination info if there are more results
    if result.next_token:
        print(
            f"\nMore results available. Use --next-token '{result.next_token}' to fetch the next page."
        )


@compute_config_cli.command(
    name="get",
    help=(
        "Get the details of a compute config.\n\n"
        "The name can contain an optional version, e.g., 'name:version'. "
        "If no version is provided, the latest one will be returned.\n\n"
    ),
    cls=AnyscaleCommand,
    example=command_examples.COMPUTE_CONFIG_GET_EXAMPLE,
)
@click.argument("compute-config-name", required=False)
@click.option(
    "-n", "--name", required=False, default=None, help="Name of the compute config.",
)
@click.option(
    "--compute-config-id",
    "--id",
    required=False,
    default=None,
    help="ID of the compute config. Alternative to name.",
    hidden=True,
)
@click.option(
    "--include-archived", is_flag=True, help="Include archived compute configurations.",
)
@click.option(
    "--cloud-id",
    required=False,
    default=None,
    type=str,
    help="Filter by cloud ID when resolving compute config by name.",
)
@click.option(
    "--cloud-name",
    required=False,
    default=None,
    type=str,
    help="Filter by cloud name when resolving compute config by name.",
)
@click.option(
    "--old-format",
    is_flag=True,
    default=False,
    help="Output the config in the old format: https://docs.anyscale.com/reference/python-sdk/models#createclustercomputeconfig.",
)
def get_compute_config(
    name: Optional[str],
    compute_config_name: Optional[str],
    compute_config_id: Optional[str],
    include_archived: bool,
    cloud_id: Optional[str],
    cloud_name: Optional[str],
    old_format: bool,
):
    """Get details of a specific compute configuration."""
    # Validate mutual exclusion: cloud_id and cloud_name cannot be used together
    if cloud_id and cloud_name:
        raise click.ClickException(
            "Error: --cloud-id and --cloud-name are mutually exclusive. "
            "Please provide only one."
        )

    name, cc_id = _validate_name_and_id_args(
        positional_name=compute_config_name, flag_name=name, id_flag=compute_config_id
    )

    # Resolve cloud filtering parameter
    cloud_filter = None
    if cloud_name:
        cloud_filter = cloud_name
    elif cloud_id:
        # Resolve cloud_id to cloud_name for the SDK
        _, cloud_filter = get_cloud_id_and_name(
            api_client=BaseController().api_client, cloud_id=cloud_id, cloud_name=None
        )

    if old_format:
        ComputeConfigController().get(
            cluster_compute_name=name,
            cluster_compute_id=cc_id,
            include_archived=include_archived,
            cloud_id=cloud_id,
            cloud_name=cloud_name,
        )
    else:
        # New format (YAML) - now supports cloud filtering
        config: ComputeConfigVersion = anyscale.compute_config.get(
            name=name, _id=cc_id, cloud=cloud_filter, include_archived=include_archived,
        )
        stream = StringIO()
        yaml.dump(config.to_dict(), stream, sort_keys=False)
        print(stream.getvalue(), end="")



import copy
import os
from string import Template
import time
from typing import Any, Dict, List, Optional

from click import ClickException
from google.api_core.exceptions import BadRequest, Forbidden, NotFound, PermissionDenied
from google.cloud.compute_v1.types import (
    FirewallPolicyAssociation,
    FirewallPolicyRule,
    Operation,
)
from google.iam.v1.iam_policy_pb2 import SetIamPolicyRequest
from google.iam.v1.policy_pb2 import Binding
from googleapiclient.errors import HttpError
import yaml

from anyscale.anyscale_pydantic import BaseModel
from anyscale.cli_logger import CloudSetupLogger
from anyscale.client.openapi_client.models import CloudAnalyticsEventCloudResource
import anyscale.conf
import anyscale.shared_anyscale_utils.conf as shared_anyscale_conf
from anyscale.util import (
    confirm,
    GCP_DEPLOYMENT_MANAGER_TIMEOUT_SECONDS_LONG,
    SharedStorageType,
)
from anyscale.utils.gcp_utils import GCP_REQUIRED_APIS, GoogleCloudClientFactory


GCP_DEPLOYMENT_PREVIEW_TIMEOUT_SECONDS = 300  # 5 minutes


GCP_MEMORYSTORE_RESOURCE_CONFIG_TEMPLATE = """
- name: redis-${CLOUD_ID}
  type: gcp-types/redis-v1:projects.locations.instances
  properties:
    parent: projects/${PROJECT_ID}/locations/${REGION}
    instanceId: redis-${CLOUD_ID}
    authorizedNetwork: projects/${PROJECT_ID}/global/networks/$$(ref.vpc-${CLOUD_ID}.name)
    memorySizeGb: 5
    replicaCount: 1
    readReplicasMode: READ_REPLICAS_ENABLED
    tier: STANDARD_HA
    redisVersion: REDIS_7_0
    displayName: redis-${CLOUD_ID}
    redisConfigs:
        maxmemory-policy: allkeys-lru
  metadata:
   dependsOn:
     - subnet-${CLOUD_ID}"""


GCP_FILESTORE_RESOURCE_CONFIG_TEMPLATE = """
- name: filestore-${CLOUD_ID}
  type: gcp-types/file-v1beta1:projects.locations.instances
  properties:
    instanceId: filestore-${CLOUD_ID}
    parent: projects/${PROJECT_ID}/locations/${REGION}-${ZONE}
    tier: STANDARD
    networks:
      - network: projects/${PROJECT_ID}/global/networks/$$(ref.vpc-${CLOUD_ID}.name)
    fileShares:
      - name: anyscale_vol
        capacityGb: ${FILESHARE_CAPACITY_GB}
    labels:
      anyscale-cloud-id: ${CLOUD_ID_UNDERSCORE}
  metadata:
    dependsOn:
      - vpc-${CLOUD_ID}"""


GCP_FILESTORE_RESOURCE_CONFIG_TEMPLATE_OA = """
- name: filestore-${CLOUD_ID}
  type: gcp-types/file-v1beta1:projects.locations.instances
  properties:
    instanceId: filestore-${CLOUD_ID}
    parent: projects/${PROJECT_ID}/locations/${REGION}-${ZONE}
    tier: STANDARD
    networks:
      - network: projects/${PROJECT_ID}/global/networks/${VPC_NAME}
    fileShares:
      - name: anyscale_vol
        capacityGb: ${FILESHARE_CAPACITY_GB}
    labels:
      anyscale-cloud-id: ${CLOUD_ID_UNDERSCORE}
- name: allow-filestore-${CLOUD_ID}
  type: compute.v1.firewall
  properties:
    direction: EGRESS
    network: projects/${PROJECT_ID}/global/networks/${VPC_NAME}
    allowed:
    - IPProtocol: all
    targetServiceAccounts:
    - $$(ref.${CLOUD_ID}.email)
    destinationRanges:
    - $$(ref.filestore-${CLOUD_ID}.networks[0].reservedIpRange)
    priority: 1000
  metadata:
    dependsOn:
      - ${CLOUD_ID}
      - filestore-${CLOUD_ID}"""


class GCPDeployment(BaseModel):
    deployment_name: str
    fingerprint: str
    config_content: str


def get_project_number(factory: GoogleCloudClientFactory, project_id: str):
    project_client = factory.resourcemanager_v3.ProjectsClient()
    try:
        project = project_client.get_project(name=f"projects/{project_id}")
        return project.name  # format: "/projects/{project_number}"
    except (NotFound, PermissionDenied, Forbidden) as e:
        raise ClickException(
            f"Error occurred when trying to access the project {project_id}: {e}"
        )


def append_project_iam_policy(
    factory: GoogleCloudClientFactory, project_id: str, role: str, member: str
):
    policy_client = factory.resourcemanager_v3.ProjectsClient()
    try:
        policy = policy_client.get_iam_policy(resource=f"projects/{project_id}")
        if policy.bindings is None:
            policy.bindings = []

        policy.bindings.append(Binding(role=role, members=[member]))
        updated_policy = policy_client.set_iam_policy(
            SetIamPolicyRequest(policy=policy, resource=f"projects/{project_id}")
        )
        return updated_policy
    except (BadRequest, NotFound, PermissionDenied, Forbidden) as e:
        raise ClickException(f"Failed to set IAM policy for project {project_id}: {e}")


def enable_project_apis(
    factory: GoogleCloudClientFactory,
    project_id: str,
    logger: CloudSetupLogger,
    enable_head_node_fault_tolerance: bool = False,
):
    """ Automatically enable APIs that Anyscale needs.
    """
    try:
        service_usage_client = factory.build("serviceusage", "v1")
        api_list = copy.deepcopy(GCP_REQUIRED_APIS)
        if enable_head_node_fault_tolerance:
            api_list.append("redis.googleapis.com")

        response = (
            service_usage_client.services()
            .batchEnable(
                parent="projects/" + project_id, body={"serviceIds": api_list},
            )
            .execute()
        )

        wait_for_operation_completion(
            service_usage_client,
            {"name": response["name"]},
            f"Enable APIs for project {project_id}",
        )
    except HttpError as e:
        logger.log_resource_exception(CloudAnalyticsEventCloudResource.GCP_PROJECT, e)
        raise ClickException(
            f"Failed to enable APIs for project {project_id}: {e}. "
            f"Please make sure the service usage API is enabled: https://console.cloud.google.com/marketplace/product/google/serviceusage.googleapis.com?project={project_id}"
        )


def get_workload_identity_pool(
    factory: GoogleCloudClientFactory, project_id: str, pool_id: str,
):
    workload_identity_pool_client = (
        factory.build("iam", "v1").projects().locations().workloadIdentityPools()
    )

    try:
        workload_identity_pool_client.get(
            name=f"projects/{project_id}/locations/global/workloadIdentityPools/{pool_id}"
        ).execute()
        return pool_id
    except HttpError as e:
        if e.status_code == 404:
            # workload identity pool not found
            return None
        else:
            raise ClickException(f"Failed to get Workload Identity Provider Pool. {e}")


def get_anyscale_gcp_access_service_acount(
    factory: GoogleCloudClientFactory, anyscale_access_service_account: str,
):
    service_account_client = factory.build("iam", "v1").projects().serviceAccounts()

    try:
        service_account_client.get(
            name=f"projects/-/serviceAccounts/{anyscale_access_service_account}"
        ).execute()
        return anyscale_access_service_account
    except HttpError as e:
        if e.status_code == 404:
            # service account not found
            return None
        else:
            raise ClickException(f"Failed to get service account: {e}")


def get_deployment_resources(
    factory: GoogleCloudClientFactory,
    deployment_name: str,
    project_id: str,
    anyscale_access_service_account_name: str,
) -> Dict[str, str]:
    """
    Get the resources in a deployment.
    """
    gcp_deployment = get_deployment_config(factory, project_id, deployment_name)
    config = yaml.safe_load(gcp_deployment.config_content)

    # TODO (congding): use a pydantic model instead of dict
    cloud_resources = {}
    for resource in config["resources"]:
        resource_props = resource["properties"]
        if resource["name"] == anyscale_access_service_account_name:
            # we should just keep the cluster service account and skip this
            continue
        if "iamMemberBinding" in resource["type"]:
            # skip IAM member binding
            continue
        if resource["type"] == "gcp-types/file-v1beta1:projects.locations.instances":
            # get filestore location and instance
            cloud_resources["filestore_instance"] = resource_props["instanceId"]
            cloud_resources["filestore_location"] = resource_props["parent"].split("/")[
                -1
            ]
        if resource["type"] == "gcp-types/redis-v1:projects.locations.instances":
            # get redis instance
            cloud_resources[
                "memorystore_name"
            ] = f'{resource_props["parent"]}/instances/{resource_props["instanceId"]}'

        cloud_resources[resource["type"]] = resource["name"]

    return cloud_resources


def create_workload_identity_pool(
    factory: GoogleCloudClientFactory,
    project_id: str,
    pool_id: str,
    logger: CloudSetupLogger,
    display_name: str = "a workload identity pool",
    description: str = "a workload identity pool",
):
    """ Create a GCP Workload Identity Provider Pool. The functionality is not
    currently supported by GCP Deployment Manager.
    """
    workload_identity_pool_client = (
        factory.build("iam", "v1").projects().locations().workloadIdentityPools()
    )

    parent = f"projects/{project_id}/locations/global"
    pool = {"displayName": display_name, "description": description}

    try:
        create_workload_identity_pool_operation = workload_identity_pool_client.create(
            parent=parent, workloadIdentityPoolId=pool_id, body=pool
        ).execute()

        wait_for_operation_completion(
            workload_identity_pool_client,
            {"name": create_workload_identity_pool_operation["name"]},
            description="creating workload identity provider pool",
        )

        workload_identity_pool = create_workload_identity_pool_operation["name"].split(
            "/operation"
        )[0]
        logger.info(f"Workload Identity Pool created: {workload_identity_pool}")
        return workload_identity_pool
    except HttpError as e:
        if e.status_code == 409:
            logger.error(
                f"Provider Pool {pool_id} already exists in project {project_id}."
            )
        else:
            logger.error(
                f"Error occurred when trying to build Workload Identity Provider Pool. Detailed: {e}"
            )
        raise ClickException("Failed to create Workload Identity Provider Pool. ")


def create_anyscale_aws_provider(
    factory: GoogleCloudClientFactory,
    organization_id: str,
    pool_id: str,
    provider_id: str,
    aws_account_id: str,
    display_name: str,
    logger: CloudSetupLogger,
):
    """ Create a GCP Workload Identity Provider for Anyscale cross account access.
    The functionality is notcurrently supported by GCP Deployment Manager.
    """
    provider_client = (
        factory.build("iam", "v1")
        .projects()
        .locations()
        .workloadIdentityPools()
        .providers()
    )

    parent = pool_id
    provider = {
        "aws": {"accountId": aws_account_id},
        "name": provider_id,
        "displayName": display_name,
        "description": "provider for Anyscale access",
        "attributeMapping": {
            "attribute.aws_role": "assertion.arn.contains('assumed-role') ? assertion.arn.extract('{account_arn}assumed-role/') + 'assumed-role/' + assertion.arn.extract('assumed-role/{role_name}/') : assertion.arn",
            "google.subject": "assertion.arn",
            "attribute.arn": "assertion.arn",
        },
        "attributeCondition": f"google.subject.startsWith('arn:aws:sts::{aws_account_id}:assumed-role/gcp_if_{organization_id}')",
    }

    try:
        response = provider_client.create(
            parent=parent, workloadIdentityPoolProviderId=provider_id, body=provider
        ).execute()
        wait_for_operation_completion(
            provider_client,
            {"name": response["name"]},
            description="creating workload identity provider",
        )
        workload_identity_provider = response["name"].split("/operation")[0]
        logger.info(f"Anyscale provider created: {workload_identity_provider}")
        return workload_identity_provider
    except HttpError as e:
        if e.status_code == 409:
            logger.error(f"Provider {provider_id} already exists in pool {parent}.")
        else:
            logger.error(
                f"Error occurred when trying to build Workload Identity Provider Pool. Detailed: {e}"
            )
        raise ClickException("Failed to create Anyscale AWS Workload Identity Provider")


def generate_deployment_manager_config(  # noqa: PLR0913
    region: str,
    project_id: str,
    cloud_id: str,
    anyscale_access_service_account_name: str,
    workload_identity_pool_name: str,
    anyscale_aws_account: str,
    organization_id: str,
    enable_head_node_fault_tolerance: bool,
    *,
    vpc_name: Optional[str] = None,
    subnet_cidr: str = "10.0.0.0/20",
    zone: str = "b",  # some regions like us-east1 or europe-west1 don't have zone "a"
    fileshare_capacity_gb: int = 1024,
    control_plane_service_account_email: Optional[str] = None,
    shared_storage: SharedStorageType = SharedStorageType.OBJECT_STORAGE,
) -> str:
    """
    Generate the deployment manager config for Anyscale cloud setup.
    """
    file_path = (
        os.path.join(anyscale.conf.ROOT_DIR_PATH, "anyscale-cloud-setup-gcp.yaml")
        if not vpc_name
        else os.path.join(
            anyscale.conf.ROOT_DIR_PATH, "anyscale-cloud-setup-gcp-oa.yaml"
        )
    )
    with open(file_path) as f:
        body = f.read()

    if enable_head_node_fault_tolerance:
        body += GCP_MEMORYSTORE_RESOURCE_CONFIG_TEMPLATE

    if shared_storage == SharedStorageType.NFS:
        if vpc_name:
            # Organization admin template - use OA filestore template
            body += GCP_FILESTORE_RESOURCE_CONFIG_TEMPLATE_OA
        else:
            # Regular template - use regular filestore template
            body += GCP_FILESTORE_RESOURCE_CONFIG_TEMPLATE

    deployment_manager_config = Template(body)

    params = {
        "ANYSCALE_HOST": shared_anyscale_conf.ANYSCALE_HOST,
        "ANYSCALE_CORS_ORIGIN": shared_anyscale_conf.ANYSCALE_CORS_ORIGIN,
        "REGION": region,
        # `cloud_id` is used as the deployment name, which cannot contain underscores.
        "CLOUD_ID": cloud_id.replace("_", "-").lower(),
        # `cloud_id_underscore` is used as the the label values, which should be consistent with GCE instances.
        "CLOUD_ID_UNDERSCORE": cloud_id,
        "PROJECT_ID": project_id,
        "ANYSCALE_ACCESS_SERVICE_ACCOUNT": anyscale_access_service_account_name,
        "WORKLOAD_IDENTITY_POOL_NAME": workload_identity_pool_name,
        "ANYSCALE_AWS_ACCOUNT": anyscale_aws_account,
        "ORGANIZATION_ID": organization_id,
        "SUBNET_CIDR": subnet_cidr,
        "ZONE": zone,
        "FILESHARE_CAPACITY_GB": fileshare_capacity_gb,
        "VPC_NAME": vpc_name,
        "ANYSCALE_CONTROL_PLANE_SERVICE_ACCOUNT_EMAIL": control_plane_service_account_email,
    }

    return deployment_manager_config.substitute(params)


def configure_firewall_policy(
    factory: GoogleCloudClientFactory,
    vpc_name: str,
    project_id: str,
    firewall_policy: str,
    subnet_cidr: str = "10.0.0.0/20",
):
    """ Add VPC association and necessary firewall rules to a firewall policy.
    """
    association = FirewallPolicyAssociation(
        name=f"{firewall_policy}-for-{vpc_name}",
        attachment_target=f"projects/{project_id}/global/networks/{vpc_name}",
    )

    rules = [
        FirewallPolicyRule(
            action="allow",
            direction="INGRESS",
            priority=1000,
            match={
                "src_ip_ranges": ["0.0.0.0/0"],
                "layer4_configs": [{"ip_protocol": "TCP", "ports": ["22", "443"]}],
            },
        ),
        FirewallPolicyRule(
            action="allow",
            direction="INGRESS",
            priority=1001,
            match={
                "src_ip_ranges": [subnet_cidr],
                "layer4_configs": [{"ip_protocol": "all"}],
            },
        ),
    ]

    network_firewall_policy_client = factory.compute_v1.NetworkFirewallPoliciesClient()
    client = factory.compute_v1.GlobalOperationsClient()
    try:
        operations = []
        for rule in rules:
            operation = network_firewall_policy_client.add_rule(
                project=project_id,
                firewall_policy=firewall_policy,
                firewall_policy_rule_resource=rule,
            )
            operations.append(operation)
        operation = network_firewall_policy_client.add_association(
            project=project_id,
            firewall_policy=firewall_policy,
            firewall_policy_association_resource=association,
        )
        operations.append(operation)

        for operation in operations:
            response = client.wait(project=project_id, operation=operation.name)
            if response.status != Operation.Status.DONE:
                # timeout
                raise ClickException(
                    "Timeout when trying to configure firewall policy."
                )
            if response.error:
                raise ClickException(
                    f"Failed to configure firewall policy {firewall_policy}. {response.error}"
                )
    except (NotFound, BadRequest) as e:
        raise ClickException(
            f"Failed to configure firewall policy {firewall_policy}. {e}"
        )


def delete_workload_identity_pool(
    factory: GoogleCloudClientFactory, pool_name: str, logger: CloudSetupLogger
):
    service = factory.build("iam", "v1").projects().locations().workloadIdentityPools()

    # we can directly delete the pool even if there're providers in it
    try:
        delete_pool_operation = service.delete(name=pool_name).execute()
        wait_for_operation_completion(
            service,
            {"name": delete_pool_operation["name"]},
            "deleting workload identity provider pool",
        )
        logger.info(f"Deleted workload identity pool: {pool_name}")
    except HttpError as e:
        if e.status_code == 404:
            return
        raise ClickException(
            f"Error occurred when trying to delete workload identity pool {pool_name}: {e}. Please delete the resources by yourself."
        )


def wait_for_operation_completion(
    service: Any,
    request: Dict[str, str],
    description: str = "Operation",
    timeout: int = 300,
    polling_interval: int = 3,
) -> None:
    start_time = time.time()
    while time.time() - start_time < timeout:
        current_operation = service.operations().get(**request).execute()
        if (
            current_operation.get("done", False)
            or current_operation.get("status", None) == "DONE"
        ):
            if "error" in current_operation:
                raise ClickException(
                    f"{description} encountered an error: {current_operation['error']}"
                )
            break
        time.sleep(polling_interval)
    else:
        raise ClickException(
            f"{description} did not complete within the timeout period ({timeout}s)"
        )


def delete_gcp_deployment(
    factory: GoogleCloudClientFactory, project_id: str, deployment_name: str,
):
    """
    Get the GCP Deployment and try to delete the deployment if it exists.
    """
    deployment_client = factory.build("deploymentmanager", "v2")
    try:
        # get the deployment
        deployment_client.deployments().get(
            project=project_id, deployment=deployment_name
        ).execute()
    except HttpError as e:
        if e.status_code == 404:
            # no deployment found
            return
        raise ClickException(
            f"Failed to get deployment {deployment_name}: {e}. Please delete the resources yourself"
        )

    try:
        # delete the deployment
        response = (
            deployment_client.deployments()
            .delete(project=project_id, deployment=deployment_name)
            .execute()
        )
        wait_for_operation_completion(
            deployment_client,
            {"operation": response["name"], "project": project_id},
            f"Delete deployment {deployment_name}",
            timeout=600,
        )
    except HttpError as e:
        raise ClickException(
            f"Failed to delete deployment {deployment_name}: {e}. Please delete the resources yourself."
        )


def update_deployment_with_bucket_only(
    factory: GoogleCloudClientFactory, project_id: str, deployment_name: str,
):
    """
    Update all resources in the deployment except the bucket.
    Removing the resources in deployment deletes the underlying resource by default.
    See https://cloud.google.com/deployment-manager/docs/reference/latest/deployments/update
    """
    # update the deployment with only the bucket
    gcp_deployment = get_deployment_config(factory, project_id, deployment_name)
    config = yaml.safe_load(gcp_deployment.config_content)

    # build a resource config with only the bucket
    resources = list(
        filter(lambda r: r["type"] == "storage.v1.bucket", config["resources"])
    )

    # update the deployment
    updated_config_content = yaml.dump({"resources": resources})
    update_deployment(
        factory,
        project_id,
        deployment_name,
        gcp_deployment.fingerprint,
        updated_config_content,
    )


def remove_firewall_policy_associations(
    factory: GoogleCloudClientFactory, project_id: str, firewall_policy_name: str,
):
    """
    Remove the firewall policy associations if exist
    """
    # get the firewall first
    network_firewall_policy_client = factory.compute_v1.NetworkFirewallPoliciesClient()
    operation_client = factory.compute_v1.GlobalOperationsClient()
    try:
        firewall_policy = network_firewall_policy_client.get(
            project=project_id, firewall_policy=firewall_policy_name
        )
    except NotFound:
        # no firewall policy
        return
    except (PermissionDenied, Forbidden, BadRequest) as e:
        raise ClickException(f"Failed to remove firewall policy associations. {e}")

    if not firewall_policy.associations:
        # no associations
        return

    operations = []
    associations = firewall_policy.associations
    for association in associations:
        try:
            remove_association_request = {
                "name": association.name,
                "project": project_id,
                "firewall_policy": firewall_policy_name,
            }
            operation = network_firewall_policy_client.remove_association(
                request=remove_association_request
            )
        except (BadRequest, NotFound):
            # BadRequest: no association found, NotFound: no firewall policy found
            continue
        except (PermissionDenied, Forbidden) as e:
            raise ClickException(f"Failed to remove firewall policy association. {e}")
        operations.append(operation)

    for operation in operations:
        response = operation_client.wait(project=project_id, operation=operation.name)
        if response.status != Operation.Status.DONE:
            # timeout
            raise ClickException(
                "Timeout when trying to remove firewall policy association."
            )
        if response.error:
            raise ClickException(
                f"Failed to remove firewall policy association. {response.error}"
            )


def delete_gcp_tls_certificates(
    factory: GoogleCloudClientFactory, project_id: str, cloud_id: str
):

    # Initialize the Compute Engine client
    certificate_manager_client = (
        factory.certificate_manager_v1.CertificateManagerClient()
    )

    try:
        location_path = certificate_manager_client.common_location_path(
            project_id, "global"
        )
    except NotFound:
        # no certificate
        return
    except Exception as e:  # noqa: BLE001
        raise ClickException(f"Failed to delete tls certificate. {e}")

    try:
        certificate_objects = certificate_manager_client.list_certificates(
            parent=location_path
        )

        certificate_maps_objects = certificate_manager_client.list_certificate_maps(
            parent=location_path
        )
    except Exception as e:  # noqa: BLE001
        raise ClickException(f"Failed to delete tls certificate. {e}")

    certificates = list(certificate_objects)
    certificate_maps = list(certificate_maps_objects)

    certificates = filter_resources(certificates, cloud_id)
    certificate_maps = filter_resources(certificate_maps, cloud_id)

    all_certificate_map_entries = []
    for certificate in certificates:
        certificate_name = certificate.split("/")[-1]
        if certificate_name:
            try:
                certificate_map_entry_path = certificate_manager_client.certificate_map_path(
                    project_id, "global", certificate_name
                )
                certificate_map_entry_objects = certificate_manager_client.list_certificate_map_entries(
                    parent=certificate_map_entry_path
                )
            except NotFound:
                # no certificate_map
                continue
            except Exception as e:  # noqa: BLE001
                raise ClickException(f"Failed to delete tls certificate. {e}")

            certificate_map_entries = list(certificate_map_entry_objects)
            all_certificate_map_entries.extend(certificate_map_entries)

    all_certificate_map_entries = filter_resources(
        all_certificate_map_entries, cloud_id
    )

    wait_on_operation(
        [
            certificate_manager_client.delete_certificate_map_entry(
                name=certificate_map_entry
            )
            for certificate_map_entry in all_certificate_map_entries
        ]
    )

    wait_on_operation(
        [
            certificate_manager_client.delete_certificate(name=certificate)
            for certificate in certificates
        ]
    )

    wait_on_operation(
        [
            certificate_manager_client.delete_certificate_map(name=certificate_map)
            for certificate_map in certificate_maps
        ]
    )


def wait_on_operation(operations):
    """
    Waits until all operations are finished before returning
    """
    for operation in operations:
        try:
            operation.result()
        except Exception as e:  # noqa: BLE001
            raise ClickException(f"Failed to wait for operation to complete. {e}")


def filter_resources(resources, cloud_id) -> List[str]:
    ret = []
    for resource in resources:
        resource_labels = resource.labels
        if (
            "anyscale-cloud-id" in resource_labels
            and resource_labels["anyscale-cloud-id"] == cloud_id
        ):
            ret.append(resource.name)
    return ret


def filter_resources_on_cloud(resources, cloud_id):
    ret = []
    for resource in resources:
        if cloud_id in resource.name:
            ret.append(resource)

    return ret


def get_or_create_memorystore_gcp(
    factory: GoogleCloudClientFactory,
    cloud_id: str,
    deployment_name: str,
    project_id: str,
    region: str,
    logger: CloudSetupLogger,
    yes: bool = False,
) -> str:
    """
    Get or create Memorystore instance in the deployment.
    Return the memorystore name in the format of projects/<project number>/locations/<location>/instances/<instance id>
    """
    redis_resource_name = f"redis-{deployment_name}"

    # Get redis from deployment
    redis_instance = get_deployment_resource(
        factory, project_id, deployment_name, redis_resource_name
    )
    if redis_instance is not None:
        return f"projects/{project_id}/locations/{region}/instances/{redis_instance['name']}"

    # No redis instance found, create one
    gcp_deployment = get_deployment_config(factory, project_id, deployment_name)

    # Update body with Memorystore
    params = {
        "CLOUD_ID": cloud_id.replace("_", "-").lower(),
        "REGION": region,
        "PROJECT_ID": project_id,
    }
    gcp_memorystore_instance = Template(
        GCP_MEMORYSTORE_RESOURCE_CONFIG_TEMPLATE
    ).substitute(params)
    updated_config_content = gcp_deployment.config_content + gcp_memorystore_instance

    # update the deployment
    if not yes:
        # Preview the deployment update
        with logger.spinner("Creating deployment preview..."):
            update_deployment(
                factory,
                project_id,
                deployment_name,
                gcp_deployment.fingerprint,
                updated_config_content,
                preview=True,
            )
        deployment_url = f"https://console.cloud.google.com/dm/deployments/details/{deployment_name}?project={project_id}"
        logger.info(f"Please review the changes in the deployment at {deployment_url}")
        confirm("Continue?", yes)

    with logger.spinner("Updating deployment..."):
        # Get the fingerprint
        gcp_deployment = get_deployment_config(factory, project_id, deployment_name)
        # To update a deployment in preview, we must set the config_content to None
        update_deployment(
            factory, project_id, deployment_name, gcp_deployment.fingerprint, None
        )

    # Get redis from deployment
    redis_instance = get_deployment_resource(
        factory, project_id, deployment_name, redis_resource_name
    )
    if redis_instance is not None:
        return f"projects/{project_id}/locations/{region}/instances/{redis_instance['name']}"
    else:
        raise ClickException(
            f"Failed to create Memorystore instance in deployment {deployment_name}. Please contact Anyscale support."
        )


def update_deployment(
    factory: GoogleCloudClientFactory,
    project_id: str,
    deployment_name: str,
    fingerprint: str,
    config_content: Optional[str],
    preview: bool = False,
):
    """
    Update a deployment with the given config.
    """
    deployment_client = factory.build("deploymentmanager", "v2")
    description = "create preview for deployment" if preview else "update deployment"
    deployment: Dict[str, Any] = {
        "name": deployment_name,
        "fingerprint": fingerprint,
    }
    if config_content:
        deployment["target"] = {
            "config": {"content": config_content},
        }

    try:
        response = (
            deployment_client.deployments()
            .update(
                project=project_id,
                deployment=deployment_name,
                body=deployment,
                preview=preview,
            )
            .execute()
        )
        wait_for_operation_completion(
            deployment_client,
            {"operation": response["name"], "project": project_id},
            f"{description.capitalize()} {deployment_name}",
            timeout=GCP_DEPLOYMENT_PREVIEW_TIMEOUT_SECONDS
            if preview
            else GCP_DEPLOYMENT_MANAGER_TIMEOUT_SECONDS_LONG,
        )
    except HttpError as e:
        raise ClickException(f"Failed to {description} {deployment_name}. Error: {e}")


def get_deployment_resource(
    factory: GoogleCloudClientFactory,
    project_id: str,
    deployment_name: str,
    resource_name: str,
):
    """
    Get a resource from a deployment. If the resource is not found, return None.
    """
    deployment_client = factory.build("deploymentmanager", "v2")
    try:
        resource = (
            deployment_client.resources()
            .get(project=project_id, deployment=deployment_name, resource=resource_name)
            .execute()
        )
        return resource
    except HttpError as e:
        if e.status_code == 404:
            return None
        raise ClickException(
            f"Failed to get resource {resource_name} from deployment {deployment_name}: {e}. "
        )


def get_deployment_config(
    factory: GoogleCloudClientFactory, project_id: str, deployment_name: str
) -> GCPDeployment:
    """
    Get the fingerprint and manifest config from a deployment.
    """
    deployment_client = factory.build("deploymentmanager", "v2")
    try:
        # get the deployment
        deployment = (
            deployment_client.deployments()
            .get(project=project_id, deployment=deployment_name)
            .execute()
        )
        fingerprint = deployment["fingerprint"]
        manifest = deployment["manifest"]
        manifest_name = manifest.split("/")[-1]

        # get manifest
        manifest = (
            deployment_client.manifests()
            .get(
                project=project_id, deployment=deployment_name, manifest=manifest_name,
            )
            .execute()
        )
    except HttpError as e:
        raise ClickException(
            f"Failed to get deployment config from deployment {deployment_name}: {e}. "
        )
    return GCPDeployment(
        deployment_name=deployment_name,
        fingerprint=fingerprint,
        config_content=manifest["config"]["content"],
    )
